from __future__ import annotations

import json
import re
from pathlib import Path

from typer.testing import CliRunner

import foldermix.packer as packer_module
import foldermix.scanner as scanner_module
from foldermix import __version__
from foldermix import cli as cli_module
from foldermix.cli import app

runner = CliRunner()
_ANSI_RE = re.compile(r"\x1b\[[0-9;]*[A-Za-z]")


def _strip_ansi(text: str) -> str:
    return _ANSI_RE.sub("", text)


def test_pack_rejects_invalid_format(tmp_path: Path) -> None:
    result = runner.invoke(app, ["pack", str(tmp_path), "--format", "bad"])
    assert result.exit_code == 1
    assert "Invalid format" in result.output
    assert "md, xml, jsonl" in result.output
    assert "--help" in result.output


def test_parse_repeatable_csv_ignores_empty_values() -> None:
    assert cli_module._parse_repeatable_csv(["", "  alpha  ", ",,", "beta,gamma", "   "]) == [
        "alpha",
        "beta",
        "gamma",
    ]


def test_pack_invalid_format_fails_before_stdin_read(monkeypatch, tmp_path: Path) -> None:
    def fail_read_stdin_paths(_use_stdin: bool, _null_delimited: bool) -> list[Path] | None:
        raise AssertionError("stdin should not be read before argument validation")

    monkeypatch.setattr("foldermix.cli._read_stdin_paths", fail_read_stdin_paths)

    result = runner.invoke(app, ["pack", str(tmp_path), "--stdin", "--format", "bad"])
    assert result.exit_code == 1
    assert "Invalid format" in result.output


def test_pack_rejects_invalid_on_oversize(tmp_path: Path) -> None:
    result = runner.invoke(app, ["pack", str(tmp_path), "--on-oversize", "bad"])
    assert result.exit_code == 1
    assert "Invalid --on-oversize" in result.output
    assert "skip, truncate" in result.output
    assert "--help" in result.output


def test_pack_rejects_invalid_redact(tmp_path: Path) -> None:
    result = runner.invoke(app, ["pack", str(tmp_path), "--redact", "bad"])
    assert result.exit_code == 1
    assert "Invalid --redact" in result.output
    assert "none, emails, phones, all" in result.output
    assert "--help" in result.output


def test_pack_rejects_negative_min_line_length(tmp_path: Path) -> None:
    result = runner.invoke(app, ["pack", str(tmp_path), "--min-line-length", "-1"])
    assert result.exit_code == 1
    assert "Invalid --min-line-length" in result.output
    assert "non-negative integer" in result.output


def test_pack_rejects_invalid_policy_fail_level(tmp_path: Path) -> None:
    result = runner.invoke(app, ["pack", str(tmp_path), "--policy-fail-level", "blocker"])
    assert result.exit_code == 1
    assert "Invalid --policy-fail-level" in result.output
    assert "low, medium, high" in result.output
    assert "critical" in result.output
    assert "--help" in result.output


def test_pack_rejects_invalid_policy_output(tmp_path: Path) -> None:
    result = runner.invoke(app, ["pack", str(tmp_path), "--policy-output", "yaml"])
    assert result.exit_code == 1
    assert "Invalid --policy-output" in result.output
    assert "text, json" in result.output
    assert "--help" in result.output


def test_pack_rejects_policy_output_without_policy_dry_run(tmp_path: Path) -> None:
    result = runner.invoke(app, ["pack", str(tmp_path), "--policy-output", "json"])
    assert result.exit_code == 1
    assert "--policy-output requires --policy-dry-run" in result.output


def test_pack_rejects_policy_output_text_without_policy_dry_run(tmp_path: Path) -> None:
    result = runner.invoke(app, ["pack", str(tmp_path), "--policy-output", "text"])
    assert result.exit_code == 1
    assert "--policy-output requires --policy-dry-run" in result.output


def test_pack_rejects_combining_dry_run_and_policy_dry_run(tmp_path: Path) -> None:
    result = runner.invoke(app, ["pack", str(tmp_path), "--dry-run", "--policy-dry-run"])
    assert result.exit_code == 1
    assert "--dry-run cannot be combined with --policy-dry-run" in result.output


def test_pack_builds_config_and_calls_packer(monkeypatch, tmp_path: Path) -> None:
    captured = {}

    def fake_pack(config) -> None:
        captured["config"] = config

    monkeypatch.setattr(packer_module, "pack", fake_pack)

    result = runner.invoke(
        app,
        [
            "pack",
            str(tmp_path),
            "--format",
            "jsonl",
            "--include-ext",
            ".py,md",
            "--exclude-ext",
            ".txt",
            "--exclude-dirs",
            "node_modules,dist",
            "--on-oversize",
            "truncate",
            "--redact",
            "all",
            "--drop-line-containing",
            "generated marker,telemetry noise",
            "--drop-line-containing",
            "multi word phrase",
            "--min-line-length",
            "8",
            "--no-include-sha256",
            "--no-include-toc",
            "--pdf-ocr",
            "--pdf-ocr-strict",
            "--fail-on-policy-violation",
            "--policy-fail-level",
            "high",
            "--policy-dry-run",
            "--policy-output",
            "json",
        ],
    )

    assert result.exit_code == 0, result.output
    config = captured["config"]
    assert config.root == tmp_path
    assert config.format == "jsonl"
    assert config.include_ext == [".py", "md"]
    assert config.exclude_ext == [".txt"]
    assert config.exclude_dirs == ["node_modules", "dist"]
    assert config.on_oversize == "truncate"
    assert config.redact == "all"
    assert config.drop_line_containing == [
        "generated marker",
        "telemetry noise",
        "multi word phrase",
    ]
    assert config.min_line_length == 8
    assert config.include_sha256 is False
    assert config.include_toc is False
    assert config.pdf_ocr is True
    assert config.pdf_ocr_strict is True
    assert config.fail_on_policy_violation is True
    assert config.policy_fail_level == "high"
    assert config.policy_dry_run is True
    assert config.policy_output == "json"


def test_pack_loads_values_from_config_file(monkeypatch, tmp_path: Path) -> None:
    captured = {}

    def fake_pack(config) -> None:
        captured["config"] = config

    monkeypatch.setattr(packer_module, "pack", fake_pack)

    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                'format = "xml"',
                'include_ext = [".py", ".md"]',
                "include_sha256 = false",
                "pdf_ocr = true",
                'drop_line_containing = ["generated marker", "trace id: "]',
                "min_line_length = 6",
                'policy_pack = "legal-hold"',
                "fail_on_policy_violation = true",
                'policy_fail_level = "critical"',
                "policy_dry_run = true",
                'policy_output = "json"',
                "",
                "[[pack.policy_rules]]",
                'rule_id = "scan-large"',
                'description = "Flag large files"',
                'stage = "scan"',
                "max_size_bytes = 100",
                "",
            ]
        ),
        encoding="utf-8",
    )

    result = runner.invoke(app, ["pack", str(tmp_path), "--config", str(config_path)])

    assert result.exit_code == 0, result.output
    config = captured["config"]
    assert config.format == "xml"
    assert config.include_ext == [".py", ".md"]
    assert config.include_sha256 is False
    assert config.pdf_ocr is True
    assert config.drop_line_containing == ["generated marker", "trace id: "]
    assert config.min_line_length == 6
    assert config.policy_pack == "legal-hold"
    assert config.fail_on_policy_violation is True
    assert config.policy_fail_level == "critical"
    assert config.policy_dry_run is True
    assert config.policy_output == "json"
    assert config.policy_rules == [
        {
            "rule_id": "scan-large",
            "description": "Flag large files",
            "stage": "scan",
            "max_size_bytes": 100,
        }
    ]


def test_pack_cli_flags_override_config_values(monkeypatch, tmp_path: Path) -> None:
    captured = {}

    def fake_pack(config) -> None:
        captured["config"] = config

    monkeypatch.setattr(packer_module, "pack", fake_pack)

    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                'format = "xml"',
                "include_toc = false",
                'policy_pack = "legal-hold"',
                "fail_on_policy_violation = false",
                'policy_fail_level = "critical"',
                "",
            ]
        ),
        encoding="utf-8",
    )

    result = runner.invoke(
        app,
        [
            "pack",
            str(tmp_path),
            "--config",
            str(config_path),
            "--format",
            "jsonl",
            "--include-toc",
            "--policy-pack",
            "strict-privacy",
            "--fail-on-policy-violation",
            "--policy-fail-level",
            "high",
        ],
    )

    assert result.exit_code == 0, result.output
    config = captured["config"]
    assert config.format == "jsonl"
    assert config.include_toc is True
    assert config.policy_pack == "strict-privacy"
    assert config.fail_on_policy_violation is True
    assert config.policy_fail_level == "high"


def test_pack_reports_invalid_config(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                'workers = "many"',
                "",
            ]
        ),
        encoding="utf-8",
    )

    result = runner.invoke(app, ["pack", str(tmp_path), "--config", str(config_path)])

    assert result.exit_code == 1
    assert "Invalid config at" in result.output
    assert "workers: expected an integer" in result.output


def test_pack_applies_config_only_fields_encoding_and_line_ending(
    monkeypatch, tmp_path: Path
) -> None:
    captured = {}

    def fake_pack(config) -> None:
        captured["config"] = config

    monkeypatch.setattr(packer_module, "pack", fake_pack)

    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                'encoding = "latin-1"',
                'line_ending = "crlf"',
                "",
            ]
        ),
        encoding="utf-8",
    )

    result = runner.invoke(app, ["pack", str(tmp_path), "--config", str(config_path)])

    assert result.exit_code == 0, result.output
    config = captured["config"]
    assert config.encoding == "latin-1"
    assert config.line_ending == "crlf"


def test_pack_rejects_configured_policy_output_without_policy_dry_run(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                'policy_output = "text"',
                "",
            ]
        ),
        encoding="utf-8",
    )

    result = runner.invoke(app, ["pack", str(tmp_path), "--config", str(config_path)])
    assert result.exit_code == 1
    assert "--policy-output requires --policy-dry-run" in result.output


def test_pack_print_effective_config_outputs_sources_and_exits(monkeypatch, tmp_path: Path) -> None:
    def fail_pack(_config) -> None:
        raise AssertionError("pack() should not be called in --print-effective-config mode")

    monkeypatch.setattr(packer_module, "pack", fail_pack)

    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                'format = "xml"',
                "include_toc = false",
                "hidden = true",
                'line_ending = "crlf"',
                "",
            ]
        ),
        encoding="utf-8",
    )

    result = runner.invoke(
        app,
        [
            "pack",
            str(tmp_path),
            "--config",
            str(config_path),
            "--format",
            "jsonl",
            "--include-toc",
            "--print-effective-config",
        ],
    )

    assert result.exit_code == 0, result.output
    payload = json.loads(result.output)
    effective = payload["effective_config"]

    assert payload["command"] == "pack"
    assert payload["config_path"] == str(config_path)
    assert effective["format"]["value"] == "jsonl"
    assert effective["format"]["source"] == "cli"
    assert effective["include_toc"]["value"] is True
    assert effective["include_toc"]["source"] == "cli"
    assert effective["hidden"]["value"] is True
    assert effective["hidden"]["source"] == "config"
    assert effective["line_ending"]["value"] == "crlf"
    assert effective["line_ending"]["source"] == "config"
    assert effective["encoding"]["value"] == "utf-8"
    assert effective["encoding"]["source"] == "default"
    assert effective["pdf_ocr"]["value"] is False
    assert effective["pdf_ocr"]["source"] == "default"
    assert effective["pdf_ocr_strict"]["value"] is False
    assert effective["pdf_ocr_strict"]["source"] == "default"
    assert effective["fail_on_policy_violation"]["value"] is False
    assert effective["fail_on_policy_violation"]["source"] == "default"
    assert effective["policy_fail_level"]["value"] == "low"
    assert effective["policy_fail_level"]["source"] == "default"


def test_pack_print_effective_config_includes_pack_defaults(monkeypatch, tmp_path: Path) -> None:
    def fail_pack(_config) -> None:
        raise AssertionError("pack() should not be called in --print-effective-config mode")

    monkeypatch.setattr(packer_module, "pack", fail_pack)

    result = runner.invoke(
        app,
        [
            "pack",
            str(tmp_path),
            "--print-effective-config",
        ],
    )

    assert result.exit_code == 0, result.output
    payload = json.loads(result.output)
    effective = payload["effective_config"]

    assert effective["line_ending"]["value"] == "lf"
    assert effective["line_ending"]["source"] == "default"
    assert effective["encoding"]["value"] == "utf-8"
    assert effective["encoding"]["source"] == "default"


def test_list_print_effective_config_outputs_sources_and_exits(monkeypatch, tmp_path: Path) -> None:
    def fail_scan(_config):
        raise AssertionError("scan() should not be called in --print-effective-config mode")

    monkeypatch.setattr(scanner_module, "scan", fail_scan)

    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[list]",
                "hidden = true",
                'include_ext = [".py"]',
                "",
            ]
        ),
        encoding="utf-8",
    )

    result = runner.invoke(
        app,
        [
            "list",
            str(tmp_path),
            "--config",
            str(config_path),
            "--include-ext",
            ".md,.txt",
            "--print-effective-config",
        ],
    )

    assert result.exit_code == 0, result.output
    payload = json.loads(result.output)
    effective = payload["effective_config"]

    assert payload["command"] == "list"
    assert effective["include_ext"]["value"] == [".md", ".txt"]
    assert effective["include_ext"]["source"] == "cli"
    assert effective["hidden"]["value"] is True
    assert effective["hidden"]["source"] == "config"


def test_stats_print_effective_config_outputs_sources_and_exits(
    monkeypatch, tmp_path: Path
) -> None:
    def fail_scan(_config):
        raise AssertionError("scan() should not be called in --print-effective-config mode")

    monkeypatch.setattr(scanner_module, "scan", fail_scan)

    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[stats]",
                'include_ext = [".py"]',
                "",
            ]
        ),
        encoding="utf-8",
    )

    result = runner.invoke(
        app,
        [
            "stats",
            str(tmp_path),
            "--config",
            str(config_path),
            "--hidden",
            "--print-effective-config",
        ],
    )

    assert result.exit_code == 0, result.output
    payload = json.loads(result.output)
    effective = payload["effective_config"]

    assert payload["command"] == "stats"
    assert effective["hidden"]["value"] is True
    assert effective["hidden"]["source"] == "cli"
    assert effective["include_ext"]["value"] == [".py"]
    assert effective["include_ext"]["source"] == "config"


def test_list_shows_included_and_skipped_files(tmp_path: Path) -> None:
    (tmp_path / "keep.txt").write_text("ok")
    (tmp_path / ".hidden").write_text("secret")

    result = runner.invoke(app, ["list", str(tmp_path)])

    assert result.exit_code == 0, result.output
    assert "keep.txt" in result.output
    assert ".hidden" not in result.output
    assert "1 files would be included, 1 skipped." in result.output


def test_skiplist_shows_skipped_files_with_reason_codes(tmp_path: Path) -> None:
    (tmp_path / "keep.txt").write_text("ok")
    (tmp_path / ".hidden").write_text("secret")

    result = runner.invoke(app, ["skiplist", str(tmp_path)])

    assert result.exit_code == 0, result.output
    assert ".hidden" in result.output
    assert "SKIP_HIDDEN" in result.output
    assert "keep.txt" not in result.output
    assert "1 files would be skipped." in result.output


def test_skiplist_conversion_check_reports_missing_optional_dependency(
    monkeypatch, tmp_path: Path
) -> None:
    (tmp_path / "doc.pdf").write_text("not a real pdf", encoding="utf-8")

    class DummyRegistry:
        def get_converter(self, _ext: str):
            return None

    monkeypatch.setattr(cli_module, "_build_converter_registry", lambda: DummyRegistry())

    result = runner.invoke(
        app, ["skiplist", str(tmp_path), "--conversion-check", "--include-ext", ".pdf"]
    )

    assert result.exit_code == 0, result.output
    assert "doc.pdf" in result.output
    assert "SKIP_OPTIONAL_DEPENDENCY_MISSING" in result.output
    assert "foldermix[pdf]" in result.output


def test_skiplist_conversion_check_reports_unsupported_extensions(tmp_path: Path) -> None:
    (tmp_path / "notes").write_text("plain text without extension", encoding="utf-8")
    (tmp_path / "custom.weird").write_text("custom content", encoding="utf-8")

    result = runner.invoke(
        app,
        [
            "skiplist",
            str(tmp_path),
            "--conversion-check",
            "--stdin",
        ],
        input=f"{tmp_path / 'notes'}\n{tmp_path / 'custom.weird'}\n",
    )

    assert result.exit_code == 0, result.output
    assert "SKIP_UNSUPPORTED_EXTENSION" in result.output
    assert "notes" in result.output
    assert "custom.weird" in result.output
    assert "files without" in result.output
    assert "an extension" in result.output
    assert "extension '.weird'" in result.output


def test_skiplist_conversion_check_uses_real_converter_registry(tmp_path: Path) -> None:
    (tmp_path / "keep.txt").write_text("ok", encoding="utf-8")

    result = runner.invoke(
        app,
        ["skiplist", str(tmp_path), "--conversion-check", "--include-ext", ".txt"],
    )

    assert result.exit_code == 0, result.output
    assert "SKIP_UNSUPPORTED_EXTENSION" not in result.output
    assert "SKIP_OPTIONAL_DEPENDENCY_MISSING" not in result.output
    assert "0 files would be skipped by scanning;" in result.output
    assert "0 additional files currently lack a" in result.output
    assert "supported converter." in result.output


def test_list_discovers_default_config(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[list]",
                "hidden = true",
                "",
            ]
        ),
        encoding="utf-8",
    )
    (tmp_path / "keep.txt").write_text("ok")
    (tmp_path / ".hidden").write_text("secret")

    result = runner.invoke(app, ["list", str(tmp_path)])

    assert result.exit_code == 0, result.output
    assert "keep.txt" in result.output
    assert ".hidden" in result.output
    assert config_path.exists()


def test_list_reports_invalid_config(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[list]",
                'hidden = "yes"',
                "",
            ]
        ),
        encoding="utf-8",
    )

    result = runner.invoke(app, ["list", str(tmp_path), "--config", str(config_path)])

    assert result.exit_code == 1
    assert "Invalid config at" in result.output
    assert "hidden: expected a boolean" in result.output


def test_skiplist_reports_invalid_config(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[list]",
                'hidden = "yes"',
                "",
            ]
        ),
        encoding="utf-8",
    )

    result = runner.invoke(app, ["skiplist", str(tmp_path), "--config", str(config_path)])

    assert result.exit_code == 1
    assert "Invalid config at" in result.output
    assert "hidden: expected a boolean" in result.output


def test_skiplist_print_effective_config_outputs_sources_and_exits(
    monkeypatch, tmp_path: Path
) -> None:
    def fail_scan(_config):
        raise AssertionError("scan() should not be called in --print-effective-config mode")

    monkeypatch.setattr(scanner_module, "scan", fail_scan)

    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[list]",
                "hidden = true",
                'include_ext = [".py"]',
                "",
            ]
        ),
        encoding="utf-8",
    )

    result = runner.invoke(
        app,
        [
            "skiplist",
            str(tmp_path),
            "--config",
            str(config_path),
            "--include-ext",
            ".md,.txt",
            "--print-effective-config",
        ],
    )

    assert result.exit_code == 0, result.output
    payload = json.loads(result.output)
    effective = payload["effective_config"]

    assert payload["command"] == "skiplist"
    assert effective["include_ext"]["value"] == [".md", ".txt"]
    assert effective["include_ext"]["source"] == "cli"
    assert effective["hidden"]["value"] is True
    assert effective["hidden"]["source"] == "config"


def test_stats_prints_summary_and_extensions(tmp_path: Path) -> None:
    (tmp_path / "a.py").write_text("print('a')\n")
    (tmp_path / "b.py").write_text("print('b')\n")
    (tmp_path / "c.md").write_text("# c\n")

    result = runner.invoke(app, ["stats", str(tmp_path)])

    assert result.exit_code == 0, result.output
    assert "Included files: 3" in result.output
    assert "Skipped files:  0" in result.output
    assert ".py" in result.output
    assert ".md" in result.output


def test_stats_reports_invalid_config(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[stats]",
                "include_ext = 123",
                "",
            ]
        ),
        encoding="utf-8",
    )

    result = runner.invoke(app, ["stats", str(tmp_path), "--config", str(config_path)])

    assert result.exit_code == 1
    assert "Invalid config at" in result.output
    assert "include_ext: expected a list of strings" in result.output


def test_version_prints_package_version() -> None:
    result = runner.invoke(app, ["version"])
    assert result.exit_code == 0
    assert f"foldermix {__version__}" in result.output


# ---------------------------------------------------------------------------
# Help-text quality tests
# ---------------------------------------------------------------------------


def test_pack_help_contains_examples() -> None:
    result = runner.invoke(app, ["pack", "--help"])
    assert result.exit_code == 0
    assert "foldermix pack" in result.output
    assert "Examples:" in result.output
    # \b blocks preserve the comment text verbatim
    assert "Pack current directory to Markdown" in result.output
    assert "Dry-run" in result.output


def test_pack_help_all_options_documented(tmp_path: Path) -> None:
    result = runner.invoke(app, ["pack", "--help"])
    assert result.exit_code == 0
    output = _strip_ansi(result.output)
    # Options that previously had no help text should now show descriptions.
    # Use short substrings that fit in any column width.
    assert "symbolic" in output  # --follow-symlinks
    assert "gitignore" in output  # --respect-gitignore
    assert "convert" in output  # --continue-on-error
    assert "frontmatter" in output  # --strip-frontmatter
    assert "SHA-256" in output  # --include-sha256
    assert "table of" in output  # --include-toc
    assert "--stdin" in output
    assert "--null" in output
    assert "--policy-fail-level" in output
    assert "--policy-dry-run" in output
    assert "--policy-output" in output


def test_list_help_all_options_documented() -> None:
    result = runner.invoke(app, ["list", "--help"])
    assert result.exit_code == 0
    output = _strip_ansi(result.output)
    # Options that previously had no help text now show descriptions
    assert "--include-ext" in output
    assert "--exclude-ext" in output
    assert "hidden" in output
    assert "gitignore" in output
    assert "--stdin" in output
    assert "--null" in output
    assert "Examples:" in output


def test_skiplist_help_all_options_documented() -> None:
    result = runner.invoke(app, ["skiplist", "--help"])
    assert result.exit_code == 0
    output = _strip_ansi(result.output)
    assert "--include-ext" in output
    assert "--exclude-ext" in output
    assert "hidden" in output
    assert "gitignore" in output
    assert "--conversion-check" in output
    assert "--scan-only" in output
    assert "--stdin" in output
    assert "--null" in output
    assert "Examples:" in output


def test_stats_help_all_options_documented() -> None:
    result = runner.invoke(app, ["stats", "--help"])
    assert result.exit_code == 0
    output = _strip_ansi(result.output)
    # Options that previously had no help text now show descriptions
    assert "--include-ext" in output
    assert "hidden" in output
    assert "--stdin" in output
    assert "--null" in output
    assert "Examples:" in output


def test_root_help_lists_all_commands() -> None:
    result = runner.invoke(app, ["--help"])
    assert result.exit_code == 0
    assert "init" in result.output
    assert "pack" in result.output
    assert "list" in result.output
    assert "skiplist" in result.output
    assert "stats" in result.output
    assert "version" in result.output
    assert "foldermix COMMAND" in result.output
    assert "github.com/shaypal5/foldermix/blob/main/docs/compliance-safety.md" in result.output



from __future__ import annotations

import sys
from pathlib import Path
from types import SimpleNamespace

from foldermix import packer
from foldermix.config import PackConfig
from foldermix.converters.base import ConversionResult
from foldermix.scanner import FileRecord


class _RegistryNone:
    @staticmethod
    def get_converter(_ext: str):
        return None


class _RegistryOne:
    def __init__(self, converter) -> None:
        self.converter = converter

    def get_converter(self, _ext: str):
        return self.converter


def test_convert_record_without_converter(tmp_path: Path) -> None:
    path = tmp_path / "a.unknown"
    path.write_text("raw", encoding="utf-8")
    record = FileRecord(path=path, relpath="a.unknown", ext=".unknown", size=3, mtime=0.0)
    item = packer._convert_record(
        record, _RegistryNone(), PackConfig(root=tmp_path, include_sha256=False)
    )
    assert item.converter_name == "none"
    assert item.content == "[No converter available for .unknown]"


def test_convert_record_truncate_and_cleanup(tmp_path: Path) -> None:
    path = tmp_path / "big.txt"
    path.write_text("0123456789abcdef", encoding="utf-8")

    class _Conv:
        @staticmethod
        def convert(p: Path, encoding: str = "utf-8") -> ConversionResult:
            return ConversionResult(content=p.read_text(encoding=encoding), converter_name="fake")

    config = PackConfig(
        root=tmp_path,
        on_oversize="truncate",
        max_bytes=8,
        include_sha256=False,
    )
    record = FileRecord(path=path, relpath="big.txt", ext=".txt", size=16, mtime=0.0)
    item = packer._convert_record(record, _RegistryOne(_Conv()), config)

    assert item.truncated is True
    assert "[TRUNCATED]" in item.content
    assert not (tmp_path / "big.txt.truncated.tmp").exists()


def test_convert_record_truncate_when_converter_deletes_tmp_file(tmp_path: Path) -> None:
    path = tmp_path / "big.txt"
    path.write_text("0123456789abcdef", encoding="utf-8")

    class _Conv:
        @staticmethod
        def convert(p: Path, encoding: str = "utf-8") -> ConversionResult:
            text = p.read_text(encoding=encoding)
            p.unlink()
            return ConversionResult(content=text, converter_name="fake")

    config = PackConfig(
        root=tmp_path,
        on_oversize="truncate",
        max_bytes=8,
        include_sha256=False,
    )
    record = FileRecord(path=path, relpath="big.txt", ext=".txt", size=16, mtime=0.0)
    item = packer._convert_record(record, _RegistryOne(_Conv()), config)

    assert item.truncated is True
    assert "[TRUNCATED]" in item.content
    assert not (tmp_path / "big.txt.truncated.tmp").exists()


def test_convert_record_applies_frontmatter_redaction_and_crlf(tmp_path: Path) -> None:
    path = tmp_path / "doc.md"
    path.write_text("x", encoding="utf-8")

    class _Conv:
        @staticmethod
        def convert(_p: Path, encoding: str = "utf-8") -> ConversionResult:
            content = "---\ntitle: test\n---\nmail test@example.com\ncall +1 (555) 123-4567\n"
            return ConversionResult(content=content, converter_name="fake")

    config = PackConfig(
        root=tmp_path,
        include_sha256=False,
        strip_frontmatter=True,
        redact="all",
        line_ending="crlf",
    )
    record = FileRecord(path=path, relpath="doc.md", ext=".md", size=1, mtime=0.0)
    item = packer._convert_record(record, _RegistryOne(_Conv()), config)

    assert "title:" not in item.content
    assert "[REDACTED_EMAIL]" in item.content
    assert "[REDACTED_PHONE]" in item.content
    assert "\r\n" in item.content


def test_convert_record_drops_lines_containing_any_filter(tmp_path: Path) -> None:
    path = tmp_path / "doc.txt"
    path.write_text("x", encoding="utf-8")

    class _Conv:
        @staticmethod
        def convert(_p: Path, encoding: str = "utf-8") -> ConversionResult:
            content = (
                "keep one\ndrop generated marker line\ndrop telemetry trace id: 123\nkeep two\n"
            )
            return ConversionResult(content=content, converter_name="fake")

    config = PackConfig(
        root=tmp_path,
        include_sha256=False,
        drop_line_containing=["generated marker", "trace id: 123"],
    )
    record = FileRecord(path=path, relpath="doc.txt", ext=".txt", size=1, mtime=0.0)
    item = packer._convert_record(record, _RegistryOne(_Conv()), config)

    assert item.content == "keep one\nkeep two\n"


def test_convert_record_drops_lines_shorter_than_min_length(tmp_path: Path) -> None:
    path = tmp_path / "doc.txt"
    path.write_text("x", encoding="utf-8")

    class _Conv:
        @staticmethod
        def convert(_p: Path, encoding: str = "utf-8") -> ConversionResult:
            return ConversionResult(content="a\nabcd\nabcde\n", converter_name="fake")

    config = PackConfig(
        root=tmp_path,
        include_sha256=False,
        min_line_length=5,
    )
    record = FileRecord(path=path, relpath="doc.txt", ext=".txt", size=1, mtime=0.0)
    item = packer._convert_record(record, _RegistryOne(_Conv()), config)

    assert item.content == "abcde\n"


def test_convert_record_ignores_sha256_oserror(monkeypatch, tmp_path: Path) -> None:
    path = tmp_path / "a.txt"
    path.write_text("abc", encoding="utf-8")

    class _Conv:
        @staticmethod
        def convert(_p: Path, encoding: str = "utf-8") -> ConversionResult:
            return ConversionResult(content="ok", converter_name="fake")

    monkeypatch.setattr(packer, "sha256_file", lambda _p: (_ for _ in ()).throw(OSError("boom")))
    record = FileRecord(path=path, relpath="a.txt", ext=".txt", size=3, mtime=0.0)
    item = packer._convert_record(record, _RegistryOne(_Conv()), PackConfig(root=tmp_path))
    assert item.sha256 is None


def test_convert_record_pdf_ocr_prefers_pdf_fallback_over_markitdown(
    monkeypatch, tmp_path: Path
) -> None:
    path = tmp_path / "a.pdf"
    path.write_text("placeholder", encoding="utf-8")

    class _MarkitdownLike:
        @staticmethod
        def convert(_p: Path, encoding: str = "utf-8") -> ConversionResult:
            return ConversionResult(content="markitdown", converter_name="markitdown")

    class _Page:
        @staticmethod
        def extract_text() -> str:
            return "from pypdf"

    class _Reader:
        def __init__(self, _path: str) -> None:
            self.pages = [_Page()]

    monkeypatch.setitem(sys.modules, "pypdf", SimpleNamespace(PdfReader=_Reader))
    record = FileRecord(path=path, relpath="a.pdf", ext=".pdf", size=1, mtime=0.0)
    item = packer._convert_record(
        record,
        _RegistryOne(_MarkitdownLike()),
        PackConfig(root=tmp_path, include_sha256=False, pdf_ocr=True),
    )

    assert item.converter_name == "pypdf"
    assert "from pypdf" in item.content
    assert "markitdown" not in item.content


def test_convert_record_pdf_without_ocr_keeps_registry_converter(tmp_path: Path) -> None:
    path = tmp_path / "a.pdf"
    path.write_text("placeholder", encoding="utf-8")

    class _MarkitdownLike:
        @staticmethod
        def convert(_p: Path, encoding: str = "utf-8") -> ConversionResult:
            return ConversionResult(content="markitdown", converter_name="markitdown")

    record = FileRecord(path=path, relpath="a.pdf", ext=".pdf", size=1, mtime=0.0)
    item = packer._convert_record(
        record,
        _RegistryOne(_MarkitdownLike()),
        PackConfig(root=tmp_path, include_sha256=False, pdf_ocr=False),
    )
    assert item.converter_name == "markitdown"
    assert item.content == "markitdown"


def test_convert_record_pdf_ocr_with_missing_pypdf_keeps_registry_converter(
    monkeypatch, tmp_path: Path
) -> None:
    path = tmp_path / "a.pdf"
    path.write_text("placeholder", encoding="utf-8")

    class _MarkitdownLike:
        @staticmethod
        def convert(_p: Path, encoding: str = "utf-8") -> ConversionResult:
            return ConversionResult(content="markitdown", converter_name="markitdown")

    monkeypatch.setitem(sys.modules, "pypdf", None)
    record = FileRecord(path=path, relpath="a.pdf", ext=".pdf", size=1, mtime=0.0)
    item = packer._convert_record(
        record,
        _RegistryOne(_MarkitdownLike()),
        PackConfig(root=tmp_path, include_sha256=False, pdf_ocr=True),
    )
    assert item.converter_name == "markitdown"
    assert item.content == "markitdown"
    assert len(item.warnings) == 1
    assert "PDF OCR is enabled" in item.warnings[0]


def test_convert_record_pdf_ocr_with_missing_pypdf_and_strict_raises(
    monkeypatch, tmp_path: Path
) -> None:
    path = tmp_path / "a.pdf"
    path.write_text("placeholder", encoding="utf-8")

    class _MarkitdownLike:
        @staticmethod
        def convert(_p: Path, encoding: str = "utf-8") -> ConversionResult:
            return ConversionResult(content="markitdown", converter_name="markitdown")

    monkeypatch.setitem(sys.modules, "pypdf", None)
    record = FileRecord(path=path, relpath="a.pdf", ext=".pdf", size=1, mtime=0.0)

    try:
        packer._convert_record(
            record,
            _RegistryOne(_MarkitdownLike()),
            PackConfig(root=tmp_path, include_sha256=False, pdf_ocr=True, pdf_ocr_strict=True),
        )
    except RuntimeError as exc:
        assert "PDF OCR is enabled" in str(exc)
    else:
        raise AssertionError("expected RuntimeError when pdf_ocr_strict is enabled")


def test_convert_record_continue_on_error_preserves_prior_ocr_warning(
    monkeypatch, tmp_path: Path
) -> None:
    path = tmp_path / "a.pdf"
    path.write_text("placeholder", encoding="utf-8")

    class _FailingConverter:
        @staticmethod
        def convert(_p: Path, encoding: str = "utf-8") -> ConversionResult:
            raise RuntimeError("converter blew up")

    monkeypatch.setitem(sys.modules, "pypdf", None)
    record = FileRecord(path=path, relpath="a.pdf", ext=".pdf", size=1, mtime=0.0)
    item = packer._convert_record(
        record,
        _RegistryOne(_FailingConverter()),
        PackConfig(
            root=tmp_path,
            include_sha256=False,
            pdf_ocr=True,
            continue_on_error=True,
        ),
    )
    assert item.converter_name == "error"
    assert "Error converting file" in item.content
    assert len(item.warnings) == 2
    assert "PDF OCR is enabled" in item.warnings[0]
    assert "converter blew up" in item.warnings[1]


def test_pack_uses_progress_branch_with_tqdm(tmp_path: Path, monkeypatch) -> None:
    (tmp_path / "a.txt").write_text("ok", encoding="utf-8")
    out_path = tmp_path / "out.md"

    fake_tqdm = SimpleNamespace(tqdm=lambda iterable, **kwargs: iterable)
    monkeypatch.setitem(sys.modules, "tqdm", fake_tqdm)

    config = PackConfig(
        root=tmp_path,
        out=out_path,
        progress=True,
        include_sha256=False,
        workers=1,
    )
    packer.pack(config)
    assert out_path.exists()


def test_pack_progress_falls_back_when_tqdm_missing(tmp_path: Path, monkeypatch) -> None:
    (tmp_path / "a.txt").write_text("ok", encoding="utf-8")
    out_path = tmp_path / "out.md"

    # When tqdm import fails, pack should still succeed via non-progress path.
    monkeypatch.setitem(sys.modules, "tqdm", None)
    config = PackConfig(
        root=tmp_path,
        out=out_path,
        progress=True,
        include_sha256=False,
        workers=1,
    )
    packer.pack(config)
    assert out_path.exists()


def test_pack_generates_default_output_name_when_out_not_set(tmp_path: Path, monkeypatch) -> None:
    project = tmp_path / "project"
    project.mkdir()
    (project / "a.txt").write_text("ok", encoding="utf-8")
    monkeypatch.chdir(tmp_path)

    config = PackConfig(root=project, out=None, format="md", include_sha256=False, workers=1)
    packer.pack(config)

    generated = list(tmp_path.glob("foldermix_*.md"))
    assert len(generated) == 1
    assert "# FolderPack Context" in generated[0].read_text(encoding="utf-8")


def test_pack_output_applies_drop_line_containing(tmp_path: Path) -> None:
    source = tmp_path / "source.txt"
    source.write_text("keep\ndrop noisy line\nkeep again\n", encoding="utf-8")
    out_path = tmp_path / "out.md"

    config = PackConfig(
        root=tmp_path,
        out=out_path,
        drop_line_containing=["noisy line"],
        include_sha256=False,
        workers=1,
    )
    packer.pack(config)

    output = out_path.read_text(encoding="utf-8")
    assert "drop noisy line" not in output
    assert "keep\nkeep again" in output


def test_pack_output_applies_min_line_length(tmp_path: Path) -> None:
    source = tmp_path / "source.txt"
    source.write_text("a\nabcd\nabcde\n", encoding="utf-8")
    out_path = tmp_path / "out.md"

    config = PackConfig(
        root=tmp_path,
        out=out_path,
        min_line_length=5,
        include_sha256=False,
        workers=1,
    )
    packer.pack(config)

    output = out_path.read_text(encoding="utf-8")
    assert "\na\n" not in output
    assert "abcd\n" not in output
    assert "abcde\n" in output



from __future__ import annotations

from pathlib import Path

from foldermix.config import PackConfig
from foldermix.scanner import SkipRecord, scan


def test_basic_scan(sample_dir: Path) -> None:
    config = PackConfig(root=sample_dir)
    included, skipped = scan(config)
    relpaths = [r.relpath for r in included]
    assert "sample.txt" in relpaths
    assert "sample.md" in relpaths


def test_hidden_excluded_by_default(sample_dir: Path) -> None:
    config = PackConfig(root=sample_dir)
    included, skipped = scan(config)
    relpaths = [r.relpath for r in included]
    assert ".hidden" not in relpaths
    skip_reasons = {r.relpath: r.reason for r in skipped}
    assert ".hidden" in skip_reasons
    assert skip_reasons[".hidden"] == "hidden"


def test_hidden_included_when_enabled(sample_dir: Path) -> None:
    config = PackConfig(root=sample_dir, hidden=True)
    included, skipped = scan(config)
    relpaths = [r.relpath for r in included]
    assert ".hidden" in relpaths


def test_exclude_ext(sample_dir: Path) -> None:
    config = PackConfig(root=sample_dir, exclude_ext=[".txt"])
    included, skipped = scan(config)
    relpaths = [r.relpath for r in included]
    assert "sample.txt" not in relpaths
    assert "sample.md" in relpaths


def test_include_ext_filter(sample_dir: Path) -> None:
    config = PackConfig(root=sample_dir, include_ext=[".py"])
    included, skipped = scan(config)
    assert all(r.ext == ".py" for r in included)


def test_include_ext_with_multiple(sample_dir: Path) -> None:
    config = PackConfig(root=sample_dir, include_ext=[".txt", ".md"])
    included, skipped = scan(config)
    assert all(r.ext in (".txt", ".md") for r in included)


def test_deterministic_ordering(sample_dir: Path) -> None:
    config = PackConfig(root=sample_dir)
    included1, _ = scan(config)
    included2, _ = scan(config)
    assert [r.relpath for r in included1] == [r.relpath for r in included2]


def test_sort_order(sample_dir: Path) -> None:
    config = PackConfig(root=sample_dir, include_ext=[".txt", ".md"])
    included, _ = scan(config)
    paths = [r.relpath for r in included]
    assert paths == sorted(paths, key=str.casefold)


def test_exclude_glob(sample_dir: Path) -> None:
    config = PackConfig(root=sample_dir, exclude_glob=["*.txt"])
    included, skipped = scan(config)
    relpaths = [r.relpath for r in included]
    assert not any(r.endswith(".txt") for r in relpaths)


def test_include_glob_overrides_exclude(sample_dir: Path) -> None:
    config = PackConfig(
        root=sample_dir,
        exclude_glob=["*.txt"],
        include_glob=["sample.txt"],
    )
    included, skipped = scan(config)
    relpaths = [r.relpath for r in included]
    assert "sample.txt" in relpaths


def test_oversize_skip(sample_dir: Path) -> None:
    config = PackConfig(root=sample_dir, max_bytes=10, on_oversize="skip")
    included, skipped = scan(config)
    for r in included:
        assert r.size <= 10
    skip_reasons = {r.relpath: r.reason for r in skipped}
    oversized = [k for k, v in skip_reasons.items() if v == "oversize"]
    assert len(oversized) > 0


def test_binary_ext_excluded(sample_dir: Path) -> None:
    config = PackConfig(root=sample_dir)
    included, skipped = scan(config)
    relpaths = [r.relpath for r in included]
    assert "image.png" not in relpaths


def test_exclude_dirs(sample_dir: Path) -> None:
    config = PackConfig(root=sample_dir, exclude_dirs=["excluded_dir"])
    included, skipped = scan(config)
    relpaths = [r.relpath for r in included]
    assert not any("excluded_dir" in r for r in relpaths)


def test_nested_files_included(sample_dir: Path) -> None:
    config = PackConfig(root=sample_dir, include_ext=[".txt"])
    included, _ = scan(config)
    relpaths = [r.relpath for r in included]
    assert "nested/child.txt" in relpaths


def test_respect_gitignore(tmp_path: Path) -> None:
    (tmp_path / ".gitignore").write_text("ignored.txt\n")
    (tmp_path / "ignored.txt").write_text("I am ignored")
    (tmp_path / "included.txt").write_text("I am included")
    config = PackConfig(root=tmp_path, respect_gitignore=True)
    included, skipped = scan(config)
    relpaths = [r.relpath for r in included]
    assert "ignored.txt" not in relpaths
    assert "included.txt" in relpaths


def test_no_respect_gitignore(tmp_path: Path) -> None:
    (tmp_path / ".gitignore").write_text("included.txt\n")
    (tmp_path / "included.txt").write_text("I am not ignored when gitignore disabled")
    config = PackConfig(root=tmp_path, respect_gitignore=False)
    included, _ = scan(config)
    relpaths = [r.relpath for r in included]
    assert "included.txt" in relpaths


def test_stdin_paths_outside_root_are_skipped(tmp_path: Path) -> None:
    root = tmp_path / "root"
    root.mkdir()
    outside = tmp_path / "outside.txt"
    outside.write_text("outside", encoding="utf-8")

    config = PackConfig(root=root, stdin_paths=[outside.resolve()])
    included, skipped = scan(config)

    assert included == []
    assert skipped == [SkipRecord("../outside.txt", "outside_root")]


def test_stdin_paths_outside_root_relpath_valueerror_fallback(tmp_path: Path, monkeypatch) -> None:
    root = tmp_path / "root"
    root.mkdir()
    outside = tmp_path / "outside.txt"
    outside.write_text("outside", encoding="utf-8")

    def raise_value_error(_path: object, _start: object = None) -> str:
        raise ValueError("cross-drive path")

    monkeypatch.setattr("foldermix.scanner.os.path.relpath", raise_value_error)

    config = PackConfig(root=root, stdin_paths=[outside.resolve()])
    included, skipped = scan(config)

    assert included == []
    assert skipped == [SkipRecord(outside.resolve().as_posix(), "outside_root")]


def test_stdin_paths_dedup_not_file_and_excluded_dir(tmp_path: Path) -> None:
    root = tmp_path / "root"
    root.mkdir()
    (root / "ok.txt").write_text("ok", encoding="utf-8")
    nested = root / "nested"
    nested.mkdir()
    excluded_dir = root / "excluded_dir"
    excluded_dir.mkdir()
    (excluded_dir / "secret.txt").write_text("secret", encoding="utf-8")

    config = PackConfig(
        root=root,
        stdin_paths=[
            nested.resolve(),  # directory should be skipped as not_file
            (excluded_dir / "secret.txt").resolve(),  # filtered by excluded_dir
            (root / "ok.txt").resolve(),
            (root / "ok.txt").resolve(),  # duplicate should be ignored
        ],
        exclude_dirs=["excluded_dir"],
    )
    included, skipped = scan(config)

    assert [record.relpath for record in included] == ["ok.txt"]
    assert skipped == [
        SkipRecord("nested", "not_file"),
        SkipRecord("excluded_dir/secret.txt", "excluded_dir"),
    ]



from __future__ import annotations

import fnmatch
import re
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from typing import Literal

PolicySeverity = Literal["low", "medium", "high", "critical"]
PolicyAction = Literal["warn", "deny"]
PolicyStage = Literal["scan", "convert", "pack", "any"]
EvaluationStage = Literal["scan", "convert", "pack"]

_VALID_SEVERITIES: set[str] = {"low", "medium", "high", "critical"}
_VALID_ACTIONS: set[str] = {"warn", "deny"}
_VALID_STAGES: set[str] = {"scan", "convert", "pack", "any"}

_MATCHER_FIELDS: set[str] = {
    "path_glob",
    "ext_in",
    "skip_reason_in",
    "content_regex",
    "max_size_bytes",
    "max_total_bytes",
    "max_file_count",
}


@dataclass(frozen=True, slots=True)
class PolicyRule:
    rule_id: str
    description: str
    severity: PolicySeverity = "medium"
    action: PolicyAction = "warn"
    stage: PolicyStage = "any"
    path_glob: str | None = None
    ext_in: tuple[str, ...] | None = None
    skip_reason_in: tuple[str, ...] | None = None
    content_regex: str | None = None
    max_size_bytes: int | None = None
    max_total_bytes: int | None = None
    max_file_count: int | None = None


@dataclass(frozen=True, slots=True)
class PolicyFinding:
    rule_id: str
    severity: PolicySeverity
    action: PolicyAction
    stage: EvaluationStage
    path: str | None
    reason_code: str
    message: str


def normalize_policy_rules(raw_rules: Sequence[Mapping[str, object]]) -> list[PolicyRule]:
    rules: list[PolicyRule] = []
    seen_rule_ids: set[str] = set()
    for idx, raw in enumerate(raw_rules):
        where = f"policy_rules[{idx}]"
        rule_id = _expect_nonempty_string(raw.get("rule_id"), f"{where}.rule_id")
        if rule_id in seen_rule_ids:
            raise ValueError(f"{where}.rule_id must be unique; duplicate value {rule_id!r}")
        seen_rule_ids.add(rule_id)
        description = _expect_nonempty_string(raw.get("description"), f"{where}.description")
        severity = _expect_literal(
            raw.get("severity", "medium"), _VALID_SEVERITIES, f"{where}.severity"
        )
        action = _expect_literal(raw.get("action", "warn"), _VALID_ACTIONS, f"{where}.action")
        stage = _expect_literal(raw.get("stage", "any"), _VALID_STAGES, f"{where}.stage")
        path_glob = _expect_optional_string(raw.get("path_glob"), f"{where}.path_glob")
        ext_in = _expect_optional_str_list(raw.get("ext_in"), f"{where}.ext_in", normalize_ext=True)
        skip_reason_in = _expect_optional_str_list(
            raw.get("skip_reason_in"), f"{where}.skip_reason_in"
        )
        content_regex = _expect_optional_string(raw.get("content_regex"), f"{where}.content_regex")
        max_size_bytes = _expect_optional_non_negative_int(
            raw.get("max_size_bytes"), f"{where}.max_size_bytes"
        )
        max_total_bytes = _expect_optional_non_negative_int(
            raw.get("max_total_bytes"), f"{where}.max_total_bytes"
        )
        max_file_count = _expect_optional_non_negative_int(
            raw.get("max_file_count"), f"{where}.max_file_count"
        )

        if content_regex is not None:
            try:
                re.compile(content_regex)
            except re.error as exc:
                raise ValueError(f"{where}.content_regex is not a valid regex: {exc}") from exc

        has_matcher = any(
            field in raw and raw[field] is not None  # type: ignore[index]
            for field in _MATCHER_FIELDS
        )
        if not has_matcher:
            raise ValueError(
                f"{where} must define at least one matcher field: "
                + ", ".join(sorted(_MATCHER_FIELDS))
            )

        rules.append(
            PolicyRule(
                rule_id=rule_id,
                description=description,
                severity=severity,  # type: ignore[arg-type]
                action=action,  # type: ignore[arg-type]
                stage=stage,  # type: ignore[arg-type]
                path_glob=path_glob,
                ext_in=ext_in,
                skip_reason_in=skip_reason_in,
                content_regex=content_regex,
                max_size_bytes=max_size_bytes,
                max_total_bytes=max_total_bytes,
                max_file_count=max_file_count,
            )
        )

    return rules


@dataclass(slots=True)
class _Event:
    stage: EvaluationStage
    path: str | None = None
    ext: str | None = None
    size_bytes: int | None = None
    skip_reason: str | None = None
    content: str | None = None
    total_bytes: int | None = None
    file_count: int | None = None


class PolicyEvaluator:
    def __init__(self, rules: Sequence[PolicyRule]) -> None:
        self._rules = sorted(
            rules,
            key=lambda rule: (rule.stage, rule.rule_id.casefold()),
        )
        self._compiled_regex: dict[str, re.Pattern[str]] = {}
        for rule in self._rules:
            if rule.content_regex is not None:
                self._compiled_regex[rule.rule_id] = re.compile(rule.content_regex)

    def evaluate_scan_included(
        self, *, path: str, ext: str, size_bytes: int
    ) -> list[PolicyFinding]:
        return self._evaluate(
            _Event(stage="scan", path=path, ext=ext, size_bytes=size_bytes),
        )

    def evaluate_scan_skipped(self, *, path: str, skip_reason: str) -> list[PolicyFinding]:
        return self._evaluate(
            _Event(stage="scan", path=path, skip_reason=skip_reason),
        )

    def evaluate_converted(
        self, *, path: str, ext: str, size_bytes: int, content: str
    ) -> list[PolicyFinding]:
        return self._evaluate(
            _Event(stage="convert", path=path, ext=ext, size_bytes=size_bytes, content=content),
        )

    def evaluate_pack_summary(self, *, file_count: int, total_bytes: int) -> list[PolicyFinding]:
        return self._evaluate(
            _Event(stage="pack", file_count=file_count, total_bytes=total_bytes),
        )

    def _evaluate(self, event: _Event) -> list[PolicyFinding]:
        findings: list[PolicyFinding] = []
        for rule in self._rules:
            if not _rule_applies_to_stage(rule, event.stage):
                continue

            reason_code = self._matches(rule, event)
            if reason_code is None:
                continue

            findings.append(
                PolicyFinding(
                    rule_id=rule.rule_id,
                    severity=rule.severity,
                    action=rule.action,
                    stage=event.stage,
                    path=event.path,
                    reason_code=reason_code,
                    message=_build_finding_message(rule, event, reason_code),
                )
            )
        return findings

    def _matches(self, rule: PolicyRule, event: _Event) -> str | None:
        matched_content_regex = False
        matched_skip_reason = False
        matched_size_limit = False
        matched_total_limit = False
        matched_count_limit = False

        if rule.path_glob is not None:
            if event.path is None or not fnmatch.fnmatch(event.path, rule.path_glob):
                return None
        if rule.ext_in is not None:
            normalized_ext = _normalize_ext(event.ext) if event.ext is not None else None
            if normalized_ext is None or normalized_ext not in rule.ext_in:
                return None
        if rule.skip_reason_in is not None:
            if event.skip_reason is None or event.skip_reason not in rule.skip_reason_in:
                return None
            matched_skip_reason = True
        if rule.content_regex is not None:
            if event.content is None:
                return None
            pattern = self._compiled_regex[rule.rule_id]
            if pattern.search(event.content) is None:
                return None
            matched_content_regex = True
        if rule.max_size_bytes is not None:
            if event.size_bytes is None or event.size_bytes <= rule.max_size_bytes:
                return None
            matched_size_limit = True
        if rule.max_total_bytes is not None:
            if event.total_bytes is None or event.total_bytes <= rule.max_total_bytes:
                return None
            matched_total_limit = True
        if rule.max_file_count is not None:
            if event.file_count is None or event.file_count <= rule.max_file_count:
                return None
            matched_count_limit = True

        if matched_total_limit:
            return "POLICY_TOTAL_BYTES_EXCEEDED"
        if matched_count_limit:
            return "POLICY_FILE_COUNT_EXCEEDED"
        if matched_size_limit:
            return "POLICY_FILE_SIZE_EXCEEDED"
        if matched_content_regex:
            return "POLICY_CONTENT_REGEX_MATCH"
        if matched_skip_reason:
            return "POLICY_SKIP_REASON_MATCH"
        return "POLICY_RULE_MATCH"


def _rule_applies_to_stage(rule: PolicyRule, stage: EvaluationStage) -> bool:
    return rule.stage == "any" or rule.stage == stage


def _build_finding_message(rule: PolicyRule, event: _Event, reason_code: str) -> str:
    if reason_code == "POLICY_FILE_SIZE_EXCEEDED" and event.size_bytes is not None:
        return f"{rule.description} (size={event.size_bytes}, max_size_bytes={rule.max_size_bytes})"
    if reason_code == "POLICY_TOTAL_BYTES_EXCEEDED" and event.total_bytes is not None:
        return (
            f"{rule.description} (total_bytes={event.total_bytes}, "
            f"max_total_bytes={rule.max_total_bytes})"
        )
    if reason_code == "POLICY_FILE_COUNT_EXCEEDED" and event.file_count is not None:
        return (
            f"{rule.description} (file_count={event.file_count}, "
            f"max_file_count={rule.max_file_count})"
        )
    if reason_code == "POLICY_SKIP_REASON_MATCH" and event.skip_reason is not None:
        return f"{rule.description} (skip_reason={event.skip_reason})"
    return rule.description


def _normalize_ext(ext: str | None) -> str | None:
    if ext is None:
        return None
    if ext == "":
        return ""
    if not ext.startswith("."):
        ext = f".{ext}"
    return ext.lower()


def _expect_nonempty_string(value: object, where: str) -> str:
    if not isinstance(value, str) or not value.strip():
        raise ValueError(f"{where} must be a non-empty string")
    return value


def _expect_optional_string(value: object, where: str) -> str | None:
    if value is None:
        return None
    if not isinstance(value, str):
        raise ValueError(f"{where} must be a string")
    return value


def _expect_literal(value: object, choices: set[str], where: str) -> str:
    if not isinstance(value, str):
        raise ValueError(f"{where} must be a string")
    if value not in choices:
        raise ValueError(f"{where} must be one of: {', '.join(sorted(choices))}")
    return value


def _expect_optional_non_negative_int(value: object, where: str) -> int | None:
    if value is None:
        return None
    if not isinstance(value, int) or isinstance(value, bool) or value < 0:
        raise ValueError(f"{where} must be a non-negative integer")
    return value


def _expect_optional_str_list(
    value: object, where: str, *, normalize_ext: bool = False
) -> tuple[str, ...] | None:
    if value is None:
        return None
    if not isinstance(value, list) or not all(isinstance(item, str) for item in value):
        raise ValueError(f"{where} must be a list of strings")
    if normalize_ext:
        normalized = tuple(_normalize_ext(item.lower()) or "" for item in value)
        return normalized
    return tuple(value)



from __future__ import annotations

import json
import shlex
import sys
from pathlib import Path

import typer
from rich.console import Console

from . import __version__
from .config import DEFAULT_EXCLUDE_DIRS, DEFAULT_EXCLUDE_EXT, PackConfig
from .config_loader import ConfigLoadError, load_command_config
from .converters.base import ConverterRegistry
from .converters.registry import build_converter_registry
from .effective_config import EffectiveConfig, effective_config_payload, merge_config_layers
from .init_profiles import available_profiles, has_profile, render_profile_config
from .report import build_skipped_file_entry
from .scanner import FileRecord, SkipRecord
from .stdin_paths import parse_stdin_paths

_INIT_PROFILE_CHOICES = ", ".join(available_profiles())

app = typer.Typer(
    name="foldermix",
    help=(
        "Pack a folder into a single LLM-friendly context file.\n\n"
        "Commands:\n\n"
        "  init    – Generate a starter foldermix.toml from a local-use profile.\n\n"
        "  pack    – Scan a directory and write all files into one output file.\n\n"
        "  list    – Preview which files would be included without packing.\n\n"
        "  skiplist – Preview which files would be skipped, with reasons.\n\n"
        "  stats   – Show file-count and byte-size statistics for a directory.\n\n"
        "  version – Print the installed foldermix version.\n\n"
        "Run 'foldermix COMMAND --help' for detailed options on any command.\n\n"
        "Guides:\n\n"
        "  Config-first workflows: https://github.com/shaypal5/foldermix/blob/main/docs/config-first-workflows.md\n\n"
        "  Compliance & safety:   https://github.com/shaypal5/foldermix/blob/main/docs/compliance-safety.md"
    ),
    add_completion=False,
)
console = Console()

_PACK_PARAM_BY_KEY = {
    "root": "path",
    "out": "out",
    "format": "format",
    "include_ext": "include_ext",
    "exclude_ext": "exclude_ext",
    "exclude_dirs": "exclude_dirs",
    "exclude_glob": "exclude_glob",
    "include_glob": "include_glob",
    "max_bytes": "max_bytes",
    "max_total_bytes": "max_total_bytes",
    "max_files": "max_files",
    "hidden": "hidden",
    "follow_symlinks": "follow_symlinks",
    "respect_gitignore": "respect_gitignore",
    "line_ending": "line_ending",
    "encoding": "encoding",
    "workers": "workers",
    "progress": "progress",
    "dry_run": "dry_run",
    "report": "report",
    "continue_on_error": "continue_on_error",
    "on_oversize": "on_oversize",
    "redact": "redact",
    "drop_line_containing": "drop_line_containing",
    "min_line_length": "min_line_length",
    "strip_frontmatter": "strip_frontmatter",
    "include_sha256": "include_sha256",
    "include_toc": "include_toc",
    "pdf_ocr": "pdf_ocr",
    "pdf_ocr_strict": "pdf_ocr_strict",
    "policy_pack": "policy_pack",
    "policy_rules": "policy_rules",
    "fail_on_policy_violation": "fail_on_policy_violation",
    "policy_fail_level": "policy_fail_level",
    "policy_dry_run": "policy_dry_run",
    "policy_output": "policy_output",
}

_LIST_PARAM_BY_KEY = {
    "root": "path",
    "include_ext": "include_ext",
    "exclude_ext": "exclude_ext",
    "hidden": "hidden",
    "respect_gitignore": "respect_gitignore",
}

_STATS_PARAM_BY_KEY = {
    "root": "path",
    "include_ext": "include_ext",
    "hidden": "hidden",
}

_OPTIONAL_CONVERTER_HINTS: dict[str, str] = {
    ".pdf": (
        "Install PDF support with 'pip install \"foldermix[pdf]\"' "
        "or OCR support with 'pip install \"foldermix[ocr]\"'."
    ),
    ".docx": "Install Office support with 'pip install \"foldermix[office]\"'.",
    ".xlsx": "Install Office support with 'pip install \"foldermix[office]\"'.",
    ".pptx": "Install Office support with 'pip install \"foldermix[office]\"'.",
}


def _parse_csv(val: str | None) -> list[str] | None:
    if val is None:
        return None
    return [v.strip() for v in val.split(",") if v.strip()]


def _parse_repeatable_csv(values: list[str] | None) -> list[str]:
    if not values:
        return []
    parsed: list[str] = []
    for raw in values:
        if "," in raw:
            parsed.extend(part.strip() for part in raw.split(",") if part.strip())
            continue
        stripped = raw.strip()
        if stripped:
            parsed.append(stripped)
    return parsed


def _print_effective_config(
    command: str, merged: EffectiveConfig, config_path: Path | None
) -> None:
    typer.echo(
        json.dumps(
            effective_config_payload(command=command, merged=merged, config_path=config_path),
            indent=2,
            sort_keys=True,
        )
    )


def _read_stdin_paths(use_stdin: bool, null_delimited: bool) -> list[Path] | None:
    if not use_stdin:
        if null_delimited:
            console.print("[red]--null requires --stdin.[/red]")
            raise typer.Exit(code=1)
        return None
    data = sys.stdin.buffer.read()
    return parse_stdin_paths(data, null_delimited=null_delimited, cwd=Path.cwd())


def _build_converter_registry() -> ConverterRegistry:
    return build_converter_registry()


def _conversion_skip_entry(record: FileRecord) -> dict[str, str]:
    ext = record.ext.lower()
    optional_hint = _OPTIONAL_CONVERTER_HINTS.get(ext)
    if optional_hint is not None:
        entry = build_skipped_file_entry(
            path=record.relpath,
            reason="optional_dependency_missing",
        )
        detail = f"No converter is available for extension {ext!r}. {optional_hint}"
        entry["message"] = f"{entry['message']} {detail}".strip()
        return entry
    if ext:
        detail = f"No converter is available for extension {ext!r} with current install."
    else:
        detail = "No converter is available for files without an extension with current install."
    entry = build_skipped_file_entry(
        path=record.relpath,
        reason="unsupported_extension",
    )
    entry["message"] = f"{entry['message']} {detail}".strip()
    return entry


def _build_skiplist_entries(
    *, included: list[FileRecord], skipped: list[SkipRecord], conversion_check: bool
) -> tuple[list[dict[str, str]], int]:
    entries = [
        build_skipped_file_entry(path=skip.relpath, reason=skip.reason)
        for skip in sorted(skipped, key=lambda record: record.relpath.casefold())
    ]
    if not conversion_check:
        return entries, 0

    registry = _build_converter_registry()
    converter_missing_count = 0
    for record in included:
        if registry.get_converter(record.ext) is None:
            entries.append(_conversion_skip_entry(record))
            converter_missing_count += 1
    entries.sort(key=lambda entry: entry["path"].casefold())
    return entries, converter_missing_count


@app.command("pack")
def pack_cmd(
    ctx: typer.Context,
    path: Path = typer.Argument(Path("."), help="Directory to pack"),
    config_path: Path | None = typer.Option(
        None, "--config", help="Path to a foldermix TOML config file"
    ),
    out: Path | None = typer.Option(
        None,
        "--out",
        "-o",
        help="Output file path (defaults to the input directory name with the appropriate extension, written next to the input directory)",
    ),
    format: str = typer.Option(
        "md", "--format", "-f", help="Output format: md, xml, jsonl [default: md]"
    ),
    include_ext: str | None = typer.Option(
        None,
        "--include-ext",
        help="Comma-separated file extensions to include (e.g. '.py,.md'). When set, only these extensions are packed.",
    ),
    exclude_ext: str | None = typer.Option(
        None,
        "--exclude-ext",
        help="Comma-separated file extensions to exclude (e.g. '.png,.zip'). Overrides the built-in exclusion list when provided.",
    ),
    exclude_dirs: str | None = typer.Option(
        None,
        "--exclude-dirs",
        help="Comma-separated directory names to exclude (e.g. 'node_modules,dist'). Overrides the built-in exclusion list when provided.",
    ),
    exclude_glob: list[str] | None = typer.Option(
        None,
        "--exclude-glob",
        help="Glob pattern(s) to exclude (e.g. '**/*.min.js'). May be repeated.",
    ),
    include_glob: list[str] | None = typer.Option(
        None,
        "--include-glob",
        help="Glob pattern(s) to include (e.g. 'src/**/*.py'). May be repeated. When set, only matching files are packed.",
    ),
    max_bytes: int = typer.Option(
        10_000_000,
        "--max-bytes",
        help="Maximum size in bytes (10 MB) for a single file. Files larger than this are handled according to --on-oversize [default: 10_000_000]",
    ),
    max_total_bytes: int | None = typer.Option(
        None,
        "--max-total-bytes",
        help="Stop packing after this many total bytes have been written. No limit by default.",
    ),
    max_files: int | None = typer.Option(
        None, "--max-files", help="Stop after packing this many files. No limit by default."
    ),
    hidden: bool = typer.Option(
        False, "--hidden", help="Include hidden files and directories (names starting with '.')"
    ),
    follow_symlinks: bool = typer.Option(
        False, "--follow-symlinks", help="Follow symbolic links when scanning the directory tree"
    ),
    respect_gitignore: bool = typer.Option(
        True,
        "--respect-gitignore/--no-respect-gitignore",
        help="Skip files listed in .gitignore [default: respect]",
    ),
    workers: int = typer.Option(
        4,
        "--workers",
        help="Number of parallel worker threads used for file conversion [default: 4]",
    ),
    progress: bool = typer.Option(
        False, "--progress", help="Show a progress bar while packing (requires the 'tqdm' package)"
    ),
    dry_run: bool = typer.Option(
        False,
        "--dry-run",
        help="List the files that would be packed without actually writing any output",
    ),
    report: Path | None = typer.Option(
        None, "--report", help="Write a JSON summary report to this path after packing"
    ),
    continue_on_error: bool = typer.Option(
        False,
        "--continue-on-error",
        help="Skip files that fail to read or convert instead of aborting",
    ),
    on_oversize: str = typer.Option(
        "skip",
        "--on-oversize",
        help="What to do when a file exceeds --max-bytes: 'skip' omits it, 'truncate' includes it up to the limit [default: skip]",
    ),
    redact: str = typer.Option(
        "none",
        "--redact",
        help="Redact sensitive patterns before writing: none, emails, phones, all [default: none]",
    ),
    drop_line_containing: list[str] | None = typer.Option(
        None,
        "--drop-line-containing",
        help=(
            "Drop lines that contain any provided literal substring. "
            "May be repeated; each value can also be comma-separated."
        ),
    ),
    min_line_length: int = typer.Option(
        0,
        "--min-line-length",
        help=(
            "Drop lines shorter than this character length. Use 0 to keep all lines [default: 0]."
        ),
    ),
    strip_frontmatter: bool = typer.Option(
        False,
        "--strip-frontmatter",
        help="Strip YAML frontmatter blocks (--- ... ---) from Markdown files before packing",
    ),
    include_sha256: bool = typer.Option(
        True,
        "--include-sha256/--no-include-sha256",
        help="Embed a SHA-256 checksum for each file in the output [default: include]",
    ),
    include_toc: bool = typer.Option(
        True,
        "--include-toc/--no-include-toc",
        help="Prepend a table of contents to Markdown output [default: include]",
    ),
    pdf_ocr: bool = typer.Option(
        False,
        "--pdf-ocr/--no-pdf-ocr",
        help="Attempt OCR for PDF pages with no extractable text when OCR dependencies are installed [default: disabled]",
    ),
    pdf_ocr_strict: bool = typer.Option(
        False,
        "--pdf-ocr-strict/--no-pdf-ocr-strict",
        help="Fail conversion when OCR is required but unavailable or unsuccessful [default: disabled]",
    ),
    policy_pack: str | None = typer.Option(
        None,
        "--policy-pack",
        help="Apply a built-in policy pack (strict-privacy, legal-hold, customer-support)",
    ),
    fail_on_policy_violation: bool = typer.Option(
        False,
        "--fail-on-policy-violation/--no-fail-on-policy-violation",
        help="Exit non-zero when policy findings meet or exceed --policy-fail-level [default: disabled]",
    ),
    policy_fail_level: str = typer.Option(
        "low",
        "--policy-fail-level",
        help="Minimum severity that triggers failure when --fail-on-policy-violation is enabled: low, medium, high, critical [default: low]",
    ),
    policy_dry_run: bool = typer.Option(
        False,
        "--policy-dry-run/--no-policy-dry-run",
        help="Run policy evaluation preview (including conversion-stage checks) without writing the packed output file",
    ),
    policy_output: str = typer.Option(
        "text",
        "--policy-output",
        help="Policy dry-run output format: text, json [default: text]",
    ),
    stdin: bool = typer.Option(
        False,
        "--stdin",
        help="Read explicit file paths from standard input instead of recursively scanning PATH.",
    ),
    null_delimited: bool = typer.Option(
        False,
        "--null",
        help="Parse stdin paths as NUL-delimited entries (compatible with find -print0). Requires --stdin.",
    ),
    print_effective_config: bool = typer.Option(
        False,
        "--print-effective-config",
        help="Print merged effective configuration with value sources and exit",
    ),
) -> None:
    """Pack a directory into a single context file.

    Recursively scans PATH (default: current directory), converts each
    eligible file to plain text, and writes the result to a single output
    file in the chosen format.

    Examples:

    \b
      # Pack current directory to Markdown (default)
      foldermix pack .

    \b
      # Pack to XML, writing to a specific file
      foldermix pack ./my-project --format xml --out context.xml

    \b
      # Include only Python and Markdown files
      foldermix pack . --include-ext .py,.md

    \b
      # Dry-run: see what would be packed without writing anything
      foldermix pack . --dry-run
    """
    from .packer import pack

    inc_ext = _parse_csv(include_ext)
    exc_ext = _parse_csv(exclude_ext) or list(DEFAULT_EXCLUDE_EXT)
    exc_dirs = _parse_csv(exclude_dirs) or list(DEFAULT_EXCLUDE_DIRS)

    values: dict[str, object] = {
        "root": path,
        "out": out,
        "format": format,
        "include_ext": inc_ext,
        "exclude_ext": exc_ext,
        "exclude_dirs": exc_dirs,
        "exclude_glob": exclude_glob or [],
        "include_glob": include_glob or [],
        "max_bytes": max_bytes,
        "max_total_bytes": max_total_bytes,
        "max_files": max_files,
        "hidden": hidden,
        "follow_symlinks": follow_symlinks,
        "respect_gitignore": respect_gitignore,
        "line_ending": "lf",
        "encoding": "utf-8",
        "workers": workers,
        "progress": progress,
        "dry_run": dry_run,
        "report": report,
        "continue_on_error": continue_on_error,
        "on_oversize": on_oversize,
        "redact": redact,
        "drop_line_containing": _parse_repeatable_csv(drop_line_containing),
        "min_line_length": min_line_length,
        "strip_frontmatter": strip_frontmatter,
        "include_sha256": include_sha256,
        "include_toc": include_toc,
        "pdf_ocr": pdf_ocr,
        "pdf_ocr_strict": pdf_ocr_strict,
        "policy_pack": policy_pack,
        "policy_rules": [],
        "fail_on_policy_violation": fail_on_policy_violation,
        "policy_fail_level": policy_fail_level,
        "policy_dry_run": policy_dry_run,
        "policy_output": policy_output,
    }

    try:
        overrides, used_config_path = load_command_config(
            "pack", root=path, config_path=config_path
        )
    except ConfigLoadError as exc:
        console.print(f"[red]{exc}[/red]")
        raise typer.Exit(code=1) from exc
    merged = merge_config_layers(
        ctx,
        defaults=values,
        config_overrides=overrides,
        key_to_param_name=_PACK_PARAM_BY_KEY,
    )
    if print_effective_config:
        _print_effective_config("pack", merged, used_config_path)
        return
    values = merged.values
    policy_output_explicitly_set = merged.sources.get("policy_output") != "default"

    if values["format"] not in ("md", "xml", "jsonl"):
        console.print(
            "[red]Invalid format:"
            f" {values['format']!r}. Valid choices are: md, xml, jsonl.[/red]\n"
            "Run 'foldermix pack --help' for full usage information."
        )
        raise typer.Exit(code=1)

    if values["on_oversize"] not in ("skip", "truncate"):
        console.print(
            "[red]Invalid --on-oversize:"
            f" {values['on_oversize']!r}. Valid choices are: skip, truncate.[/red]\n"
            "Run 'foldermix pack --help' for full usage information."
        )
        raise typer.Exit(code=1)

    if values["redact"] not in ("none", "emails", "phones", "all"):
        console.print(
            "[red]Invalid --redact:"
            f" {values['redact']!r}. Valid choices are: none, emails, phones, all.[/red]\n"
            "Run 'foldermix pack --help' for full usage information."
        )
        raise typer.Exit(code=1)

    if values["min_line_length"] < 0:
        console.print(
            "[red]Invalid --min-line-length:"
            f" {values['min_line_length']!r}. Value must be a non-negative integer.[/red]\n"
            "Run 'foldermix pack --help' for full usage information."
        )
        raise typer.Exit(code=1)

    if values["policy_fail_level"] not in ("low", "medium", "high", "critical"):
        console.print(
            "[red]Invalid --policy-fail-level:"
            f" {values['policy_fail_level']!r}. Valid choices are: low, medium, high, critical.[/red]\n"
            "Run 'foldermix pack --help' for full usage information."
        )
        raise typer.Exit(code=1)

    if values["policy_output"] not in ("text", "json"):
        console.print(
            "[red]Invalid --policy-output:"
            f" {values['policy_output']!r}. Valid choices are: text, json.[/red]\n"
            "Run 'foldermix pack --help' for full usage information."
        )
        raise typer.Exit(code=1)

    if policy_output_explicitly_set and not values["policy_dry_run"]:
        console.print(
            "[red]--policy-output requires --policy-dry-run.[/red]\n"
            "Run 'foldermix pack --help' for full usage information."
        )
        raise typer.Exit(code=1)

    if values["dry_run"] and values["policy_dry_run"]:
        console.print(
            "[red]--dry-run cannot be combined with --policy-dry-run.[/red]\n"
            "Run 'foldermix pack --help' for full usage information."
        )
        raise typer.Exit(code=1)

    stdin_paths = _read_stdin_paths(stdin, null_delimited)
    pack_config = PackConfig(stdin_paths=stdin_paths, **values)  # type: ignore[arg-type]

    pack(pack_config)


@app.command("list")
def list_cmd(
    ctx: typer.Context,
    path: Path = typer.Argument(Path("."), help="Directory to scan"),
    config_path: Path | None = typer.Option(
        None, "--config", help="Path to a foldermix TOML config file"
    ),
    include_ext: str | None = typer.Option(
        None, "--include-ext", help="Comma-separated file extensions to include (e.g. '.py,.md')"
    ),
    exclude_ext: str | None = typer.Option(
        None, "--exclude-ext", help="Comma-separated file extensions to exclude (e.g. '.png,.zip')"
    ),
    hidden: bool = typer.Option(
        False, "--hidden", help="Include hidden files and directories (names starting with '.')"
    ),
    respect_gitignore: bool = typer.Option(
        True,
        "--respect-gitignore/--no-respect-gitignore",
        help="Skip files listed in .gitignore [default: respect]",
    ),
    stdin: bool = typer.Option(
        False,
        "--stdin",
        help="Read explicit file paths from standard input instead of recursively scanning PATH.",
    ),
    null_delimited: bool = typer.Option(
        False,
        "--null",
        help="Parse stdin paths as NUL-delimited entries (compatible with find -print0). Requires --stdin.",
    ),
    print_effective_config: bool = typer.Option(
        False,
        "--print-effective-config",
        help="Print merged effective configuration with value sources and exit",
    ),
) -> None:
    """List candidate files that would be packed.

    Performs the same file-discovery logic as 'foldermix pack' but only
    prints the resulting file list — nothing is written.  Useful for
    previewing which files would be included before committing to a full
    pack run.

    Examples:

    \b
      # List all files in the current directory
      foldermix list .

    \b
      # List only Python files, including hidden ones
      foldermix list . --include-ext .py --hidden
    """
    from .config import PackConfig
    from .scanner import scan

    values: dict[str, object] = {
        "root": path,
        "include_ext": _parse_csv(include_ext),
        "exclude_ext": _parse_csv(exclude_ext) or list(DEFAULT_EXCLUDE_EXT),
        "hidden": hidden,
        "respect_gitignore": respect_gitignore,
    }
    try:
        overrides, used_config_path = load_command_config(
            "list", root=path, config_path=config_path
        )
    except ConfigLoadError as exc:
        console.print(f"[red]{exc}[/red]")
        raise typer.Exit(code=1) from exc
    merged = merge_config_layers(
        ctx,
        defaults=values,
        config_overrides=overrides,
        key_to_param_name=_LIST_PARAM_BY_KEY,
    )
    if print_effective_config:
        _print_effective_config("list", merged, used_config_path)
        return
    values = merged.values
    stdin_paths = _read_stdin_paths(stdin, null_delimited)

    pack_config = PackConfig(
        root=values["root"],  # type: ignore[arg-type]
        stdin_paths=stdin_paths,
        include_ext=values["include_ext"],  # type: ignore[arg-type]
        exclude_ext=values["exclude_ext"],  # type: ignore[arg-type]
        hidden=values["hidden"],  # type: ignore[arg-type]
        respect_gitignore=values["respect_gitignore"],  # type: ignore[arg-type]
    )
    included, skipped = scan(pack_config)
    for r in included:
        console.print(f"{r.relpath}  ({r.size:,} bytes)")
    console.print(f"\n{len(included)} files would be included, {len(skipped)} skipped.")


@app.command("stats")
def stats_cmd(
    ctx: typer.Context,
    path: Path = typer.Argument(Path("."), help="Directory to scan"),
    config_path: Path | None = typer.Option(
        None, "--config", help="Path to a foldermix TOML config file"
    ),
    include_ext: str | None = typer.Option(
        None, "--include-ext", help="Comma-separated file extensions to include (e.g. '.py,.md')"
    ),
    hidden: bool = typer.Option(
        False, "--hidden", help="Include hidden files and directories (names starting with '.')"
    ),
    stdin: bool = typer.Option(
        False,
        "--stdin",
        help="Read explicit file paths from standard input instead of recursively scanning PATH.",
    ),
    null_delimited: bool = typer.Option(
        False,
        "--null",
        help="Parse stdin paths as NUL-delimited entries (compatible with find -print0). Requires --stdin.",
    ),
    print_effective_config: bool = typer.Option(
        False,
        "--print-effective-config",
        help="Print merged effective configuration with value sources and exit",
    ),
) -> None:
    """Show summary statistics for a directory.

    Counts files and bytes broken down by file extension, giving a quick
    overview of what a directory contains before you pack it.

    Examples:

    \b
      # Stats for the current directory
      foldermix stats .

    \b
      # Stats for Python files only
      foldermix stats ./src --include-ext .py
    """
    from .config import PackConfig
    from .scanner import scan

    values: dict[str, object] = {
        "root": path,
        "include_ext": _parse_csv(include_ext),
        "hidden": hidden,
    }
    try:
        overrides, used_config_path = load_command_config(
            "stats", root=path, config_path=config_path
        )
    except ConfigLoadError as exc:
        console.print(f"[red]{exc}[/red]")
        raise typer.Exit(code=1) from exc
    merged = merge_config_layers(
        ctx,
        defaults=values,
        config_overrides=overrides,
        key_to_param_name=_STATS_PARAM_BY_KEY,
    )
    if print_effective_config:
        _print_effective_config("stats", merged, used_config_path)
        return
    values = merged.values
    stdin_paths = _read_stdin_paths(stdin, null_delimited)

    pack_config = PackConfig(
        root=values["root"],  # type: ignore[arg-type]
        stdin_paths=stdin_paths,
        include_ext=values["include_ext"],  # type: ignore[arg-type]
        hidden=values["hidden"],  # type: ignore[arg-type]
    )
    included, skipped = scan(pack_config)
    total_bytes = sum(r.size for r in included)

    ext_counts: dict[str, int] = {}
    for r in included:
        ext_counts[r.ext] = ext_counts.get(r.ext, 0) + 1

    console.print(f"[bold]Stats for {path}[/bold]")
    console.print(f"  Included files: {len(included)}")
    console.print(f"  Skipped files:  {len(skipped)}")
    console.print(f"  Total bytes:    {total_bytes:,}")
    console.print("\n  [bold]By extension:[/bold]")
    for ext, count in sorted(ext_counts.items(), key=lambda x: -x[1]):
        console.print(f"    {ext or '(none)':15s} {count:5d}")


@app.command("skiplist")
def skiplist_cmd(
    ctx: typer.Context,
    path: Path = typer.Argument(Path("."), help="Directory to scan"),
    config_path: Path | None = typer.Option(
        None, "--config", help="Path to a foldermix TOML config file"
    ),
    include_ext: str | None = typer.Option(
        None, "--include-ext", help="Comma-separated file extensions to include (e.g. '.py,.md')"
    ),
    exclude_ext: str | None = typer.Option(
        None, "--exclude-ext", help="Comma-separated file extensions to exclude (e.g. '.png,.zip')"
    ),
    hidden: bool = typer.Option(
        False, "--hidden", help="Include hidden files and directories (names starting with '.')"
    ),
    respect_gitignore: bool = typer.Option(
        True,
        "--respect-gitignore/--no-respect-gitignore",
        help="Skip files listed in .gitignore [default: respect]",
    ),
    conversion_check: bool = typer.Option(
        False,
        "--conversion-check/--scan-only",
        help=(
            "Also include files with no available converter in the current "
            "optional-dependency install [default: scan-only]"
        ),
    ),
    stdin: bool = typer.Option(
        False,
        "--stdin",
        help="Read explicit file paths from standard input instead of recursively scanning PATH.",
    ),
    null_delimited: bool = typer.Option(
        False,
        "--null",
        help="Parse stdin paths as NUL-delimited entries (compatible with find -print0). Requires --stdin.",
    ),
    print_effective_config: bool = typer.Option(
        False,
        "--print-effective-config",
        help="Print merged effective configuration with value sources and exit",
    ),
) -> None:
    """Show files that would be skipped, with reasons.

    By default, this is the inverse of `foldermix list` for scanner-level
    filtering decisions. With `--conversion-check`, it also reports files
    that have no available converter in the current install.

    Examples:

    \b
      # Show scanner-level skipped files
      foldermix skiplist .

    \b
      # Include conversion-availability checks
      foldermix skiplist . --conversion-check
    """
    from .scanner import scan

    values: dict[str, object] = {
        "root": path,
        "include_ext": _parse_csv(include_ext),
        "exclude_ext": _parse_csv(exclude_ext) or list(DEFAULT_EXCLUDE_EXT),
        "hidden": hidden,
        "respect_gitignore": respect_gitignore,
    }
    try:
        overrides, used_config_path = load_command_config(
            "list", root=path, config_path=config_path
        )
    except ConfigLoadError as exc:
        console.print(f"[red]{exc}[/red]")
        raise typer.Exit(code=1) from exc
    merged = merge_config_layers(
        ctx,
        defaults=values,
        config_overrides=overrides,
        key_to_param_name=_LIST_PARAM_BY_KEY,
    )
    if print_effective_config:
        _print_effective_config("skiplist", merged, used_config_path)
        return
    values = merged.values
    stdin_paths = _read_stdin_paths(stdin, null_delimited)

    pack_config = PackConfig(
        root=values["root"],  # type: ignore[arg-type]
        stdin_paths=stdin_paths,
        include_ext=values["include_ext"],  # type: ignore[arg-type]
        exclude_ext=values["exclude_ext"],  # type: ignore[arg-type]
        hidden=values["hidden"],  # type: ignore[arg-type]
        respect_gitignore=values["respect_gitignore"],  # type: ignore[arg-type]
    )
    included, skipped = scan(pack_config)
    skip_entries, converter_missing_count = _build_skiplist_entries(
        included=included,
        skipped=skipped,
        conversion_check=conversion_check,
    )
    for entry in skip_entries:
        console.print(
            f"{entry['path']}  [{entry['reason_code']}] {entry['message']}",
            markup=False,
        )
    if conversion_check:
        console.print(
            "\n"
            f"{len(skipped)} files would be skipped by scanning; "
            f"{converter_missing_count} additional files currently lack a supported converter."
        )
    else:
        console.print(f"\n{len(skip_entries)} files would be skipped.")


@app.command("init")
def init_cmd(
    profile: str = typer.Option(
        ...,
        "--profile",
        help=(
            f"Starter profile name: {_INIT_PROFILE_CHOICES}. "
            "Use this to bootstrap a local foldermix.toml."
        ),
    ),
    out: Path = typer.Option(
        Path("foldermix.toml"),
        "--out",
        "-o",
        help="Path to write the generated config file [default: ./foldermix.toml]",
    ),
    force: bool = typer.Option(
        False,
        "--force",
        help="Overwrite an existing output file. By default existing files are preserved.",
    ),
) -> None:
    """Generate a starter foldermix.toml for common local workflows."""
    normalized = profile.strip().lower()
    if not has_profile(normalized):
        valid = ", ".join(available_profiles())
        console.print(
            "[red]Invalid profile:"
            f" {profile!r}. Valid choices are: {valid}.[/red]\n"
            "Run 'foldermix init --help' for usage information."
        )
        raise typer.Exit(code=1)

    output_path = out.expanduser()
    if output_path.exists() and not force:
        console.print(
            f"[red]Refusing to overwrite existing file: {output_path}.[/red]\n"
            "Use --force to overwrite."
        )
        raise typer.Exit(code=1)

    try:
        output_path.parent.mkdir(parents=True, exist_ok=True)
        output_path.write_text(render_profile_config(normalized), encoding="utf-8")
    except OSError as exc:
        console.print(f"[red]Failed to write config: {exc}[/red]")
        raise typer.Exit(code=1) from exc

    console.print(
        f"Wrote starter config to {output_path} using profile '{normalized}'. "
        "Run: foldermix pack . --config "
        f"{shlex.quote(str(output_path))}"
    )


@app.command("version")
def version_cmd() -> None:
    """Print the version."""
    console.print(f"foldermix {__version__}")


if __name__ == "__main__":
    app()



from __future__ import annotations

from pathlib import Path

import pytest
from click.core import ParameterSource

from foldermix.effective_config import effective_config_payload, merge_config_layers


class FakeContext:
    def __init__(self, sources: dict[str, ParameterSource | None] | None = None) -> None:
        self._sources = sources or {}

    def get_parameter_source(self, param_name: str) -> ParameterSource | None:
        return self._sources.get(param_name, ParameterSource.DEFAULT)


@pytest.mark.parametrize(
    ("key", "default_value", "config_value", "cli_value"),
    [
        ("hidden", False, True, False),
        ("include_ext", None, [".py"], [".md"]),
        ("format", "md", "xml", "jsonl"),
    ],
)
def test_merge_config_layers_cli_values_override_config(
    key: str, default_value: object, config_value: object, cli_value: object
) -> None:
    merged = merge_config_layers(
        FakeContext({key: ParameterSource.COMMANDLINE}),
        defaults={key: cli_value},
        config_overrides={key: config_value},
    )

    assert merged.values[key] == cli_value
    assert merged.sources[key] == "cli"

    merged_no_cli = merge_config_layers(
        FakeContext(),
        defaults={key: default_value},
        config_overrides={key: config_value},
    )
    assert merged_no_cli.values[key] == config_value
    assert merged_no_cli.sources[key] == "config"


def test_merge_config_layers_adds_config_only_key() -> None:
    merged = merge_config_layers(
        FakeContext(),
        defaults={"format": "md"},
        config_overrides={"encoding": "latin-1"},
    )

    assert merged.values["format"] == "md"
    assert merged.sources["format"] == "default"
    assert merged.values["encoding"] == "latin-1"
    assert merged.sources["encoding"] == "config"


def test_merge_config_layers_honors_parameter_name_mapping() -> None:
    merged = merge_config_layers(
        FakeContext({"path": ParameterSource.COMMANDLINE}),
        defaults={"root": Path(".")},
        config_overrides={},
        key_to_param_name={"root": "path"},
    )

    assert merged.sources["root"] == "cli"


def test_effective_config_payload_is_deterministic_and_json_safe() -> None:
    merged = merge_config_layers(
        FakeContext({"path": ParameterSource.COMMANDLINE}),
        defaults={"format": "md", "root": Path("/tmp/project")},
        config_overrides={"include_ext": [".py", ".md"]},
        key_to_param_name={"root": "path"},
    )

    payload = effective_config_payload(
        command="pack", merged=merged, config_path=Path("/tmp/project/foldermix.toml")
    )

    assert payload["command"] == "pack"
    assert Path(payload["config_path"]) == Path("/tmp/project/foldermix.toml")
    assert list(payload["effective_config"]) == sorted(payload["effective_config"])
    assert Path(payload["effective_config"]["root"]["value"]) == Path("/tmp/project")
    assert payload["effective_config"]["root"]["source"] == "cli"
    assert payload["effective_config"]["include_ext"]["value"] == [".py", ".md"]


def test_effective_config_payload_converts_nested_dict_values() -> None:
    merged = merge_config_layers(
        FakeContext(),
        defaults={
            "nested": {
                1: Path("/tmp/a"),
                "items": [Path("/tmp/b"), {"k": Path("/tmp/c")}],
            }
        },
        config_overrides={},
    )

    payload = effective_config_payload(command="pack", merged=merged, config_path=None)
    nested = payload["effective_config"]["nested"]["value"]

    assert payload["config_path"] is None
    assert Path(nested["1"]) == Path("/tmp/a")
    assert Path(nested["items"][0]) == Path("/tmp/b")
    assert Path(nested["items"][1]["k"]) == Path("/tmp/c")



from __future__ import annotations

from pathlib import Path

import pytest

from foldermix import config_loader
from foldermix.config_loader import ConfigLoadError, load_command_config


def test_load_command_config_from_tool_section(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[tool.foldermix.pack]",
                'format = "xml"',
                "workers = 2",
                'include_ext = [".py", ".md"]',
                "include_toc = false",
                "pdf_ocr = true",
                "pdf_ocr_strict = false",
                "",
            ]
        ),
        encoding="utf-8",
    )

    values, used_path = load_command_config("pack", root=tmp_path, config_path=config_path)

    assert used_path == config_path
    assert values["format"] == "xml"
    assert values["workers"] == 2
    assert values["include_ext"] == [".py", ".md"]
    assert values["include_toc"] is False
    assert values["pdf_ocr"] is True
    assert values["pdf_ocr_strict"] is False


def test_load_command_config_discovers_parent_file(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[list]",
                "hidden = true",
                "",
            ]
        ),
        encoding="utf-8",
    )
    nested = tmp_path / "a" / "b"
    nested.mkdir(parents=True)

    values, used_path = load_command_config("list", root=nested, config_path=None)

    assert used_path == config_path
    assert values["hidden"] is True


def test_load_command_config_rejects_invalid_types(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                'workers = "fast"',
                "",
            ]
        ),
        encoding="utf-8",
    )

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("pack", root=tmp_path, config_path=config_path)

    message = str(exc.value)
    assert "Invalid config at" in message
    assert "workers: expected an integer" in message


def test_load_command_config_rejects_unknown_key(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[stats]",
                "unknown = 1",
                "",
            ]
        ),
        encoding="utf-8",
    )

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("stats", root=tmp_path, config_path=config_path)

    assert "unknown key" in str(exc.value)


def test_load_command_config_rejects_known_but_invalid_command_key(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[stats]",
                "workers = 2",
                "",
            ]
        ),
        encoding="utf-8",
    )

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("stats", root=tmp_path, config_path=config_path)

    assert "is not valid for 'stats' command" in str(exc.value)


def test_load_command_config_accepts_csv_list_values(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                'include_ext = ".py, .md,  .txt "',
                'drop_line_containing = "generated marker, telemetry: trace id"',
                "min_line_length = 7",
                "",
            ]
        ),
        encoding="utf-8",
    )

    values, used_path = load_command_config("pack", root=tmp_path, config_path=config_path)

    assert used_path == config_path
    assert values["include_ext"] == [".py", ".md", ".txt"]
    assert values["drop_line_containing"] == ["generated marker", "telemetry: trace id"]
    assert values["min_line_length"] == 7


def test_load_command_config_accepts_drop_line_containing_list_literals(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                'drop_line_containing = ["do not edit, generated", "multi word marker"]',
                "",
            ]
        ),
        encoding="utf-8",
    )

    values, used_path = load_command_config("pack", root=tmp_path, config_path=config_path)

    assert used_path == config_path
    assert values["drop_line_containing"] == ["do not edit, generated", "multi word marker"]


def test_load_command_config_rejects_invalid_list_shape(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                'include_ext = [".py", 1]',
                "",
            ]
        ),
        encoding="utf-8",
    )

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("pack", root=tmp_path, config_path=config_path)

    assert "expected a list of strings" in str(exc.value)


def test_load_command_config_rejects_negative_min_line_length(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                "min_line_length = -1",
                "",
            ]
        ),
        encoding="utf-8",
    )

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("pack", root=tmp_path, config_path=config_path)

    assert "min_line_length: expected a non-negative integer" in str(exc.value)


def test_load_command_config_rejects_literal_non_string(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                "format = 123",
                "",
            ]
        ),
        encoding="utf-8",
    )

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("pack", root=tmp_path, config_path=config_path)

    assert "format: expected a string" in str(exc.value)


def test_load_command_config_rejects_invalid_literal_choice(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                'format = "bad"',
                "",
            ]
        ),
        encoding="utf-8",
    )

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("pack", root=tmp_path, config_path=config_path)

    assert "expected one of" in str(exc.value)


def test_load_command_config_rejects_invalid_bool_and_string_types(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                'hidden = "yes"',
                'pdf_ocr = "yes"',
                "encoding = 1",
                "",
            ]
        ),
        encoding="utf-8",
    )

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("pack", root=tmp_path, config_path=config_path)

    message = str(exc.value)
    assert "hidden: expected a boolean" in message
    assert "pdf_ocr: expected a boolean" in message
    assert "encoding: expected a string" in message


def test_load_command_config_coerces_path_keys(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                'out = "bundle.xml"',
                "",
            ]
        ),
        encoding="utf-8",
    )

    values, _ = load_command_config("pack", root=tmp_path, config_path=config_path)
    assert values["out"] == Path("bundle.xml")


def test_load_command_config_rejects_invalid_path_type(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                "report = 123",
                "",
            ]
        ),
        encoding="utf-8",
    )

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("pack", root=tmp_path, config_path=config_path)

    assert "report: expected a filesystem path string" in str(exc.value)


def test_load_command_config_supports_flat_root_without_sections(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                'format = "xml"',
                "workers = 2",
                "",
            ]
        ),
        encoding="utf-8",
    )

    values, _ = load_command_config("pack", root=tmp_path, config_path=config_path)
    assert values["format"] == "xml"
    assert values["workers"] == 2


def test_load_command_config_rejects_unknown_top_level_keys_in_section_mode(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "extra = 1",
                "",
                "[pack]",
                'format = "md"',
                "",
            ]
        ),
        encoding="utf-8",
    )

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("pack", root=tmp_path, config_path=config_path)

    assert "unknown top-level keys" in str(exc.value)


def test_load_command_config_rejects_non_table_common_section(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "common = 1",
                "",
                "[pack]",
                'format = "md"',
                "",
            ]
        ),
        encoding="utf-8",
    )

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("pack", root=tmp_path, config_path=config_path)

    assert "config.common must be a TOML table" in str(exc.value)


def test_load_command_config_merges_common_section_and_command_override(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[common]",
                "hidden = true",
                "",
                "[pack]",
                "hidden = false",
                "workers = 2",
                "",
            ]
        ),
        encoding="utf-8",
    )

    values, _ = load_command_config("pack", root=tmp_path, config_path=config_path)
    assert values["hidden"] is False
    assert values["workers"] == 2


def test_load_command_config_uses_common_section_when_command_section_missing(
    tmp_path: Path,
) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[common]",
                "hidden = true",
                "",
            ]
        ),
        encoding="utf-8",
    )

    values, _ = load_command_config("list", root=tmp_path, config_path=config_path)
    assert values["hidden"] is True


def test_load_command_config_rejects_non_table_command_section(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "list = 1",
                "",
                "[common]",
                "hidden = true",
                "",
            ]
        ),
        encoding="utf-8",
    )

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("list", root=tmp_path, config_path=config_path)

    assert "config.list must be a TOML table" in str(exc.value)


def test_load_command_config_supports_top_level_foldermix_table(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[foldermix.pack]",
                'format = "jsonl"',
                "",
            ]
        ),
        encoding="utf-8",
    )

    values, _ = load_command_config("pack", root=tmp_path, config_path=config_path)
    assert values["format"] == "jsonl"


def test_load_command_config_rejects_non_table_tool_foldermix(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[tool]",
                "foldermix = 1",
                "",
            ]
        ),
        encoding="utf-8",
    )

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("pack", root=tmp_path, config_path=config_path)

    assert "tool.foldermix must be a TOML table" in str(exc.value)


def test_load_command_config_rejects_non_table_top_level_foldermix(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text("foldermix = 1\n", encoding="utf-8")

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("pack", root=tmp_path, config_path=config_path)

    assert "foldermix must be a TOML table" in str(exc.value)


def test_discover_config_path_handles_file_start_and_missing_path(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text("", encoding="utf-8")
    nested_dir = tmp_path / "nested"
    nested_dir.mkdir()
    target_file = nested_dir / "target.txt"
    target_file.write_text("x", encoding="utf-8")

    assert config_loader.discover_config_path(target_file) == config_path

    monkeypatch.chdir(tmp_path)
    assert config_loader.discover_config_path(tmp_path / "does-not-exist") == config_path


def test_load_command_config_rejects_unsupported_command(tmp_path: Path) -> None:
    with pytest.raises(ValueError):
        load_command_config("unknown", root=tmp_path, config_path=None)


def test_load_command_config_rejects_missing_explicit_path(tmp_path: Path) -> None:
    config_path = tmp_path / "missing.toml"
    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("pack", root=tmp_path, config_path=config_path)

    assert "config file does not exist" in str(exc.value)


def test_load_command_config_rejects_invalid_toml(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text("[pack\n", encoding="utf-8")

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("pack", root=tmp_path, config_path=config_path)

    assert "invalid TOML" in str(exc.value)


def test_load_command_config_wraps_os_errors(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text("[pack]\n", encoding="utf-8")

    def fail_read_text(self: Path, *, encoding: str) -> str:
        raise OSError("boom")

    monkeypatch.setattr(Path, "read_text", fail_read_text)

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("pack", root=tmp_path, config_path=config_path)

    assert "could not read config" in str(exc.value)


def test_load_command_config_rejects_non_table_after_parse(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text("[pack]\n", encoding="utf-8")

    monkeypatch.setattr(config_loader.tomllib, "loads", lambda _: ["not-a-table"])

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("pack", root=tmp_path, config_path=config_path)

    assert "config root must be a TOML table" in str(exc.value)


def test_private_coercion_helpers_cover_optional_and_unsupported_keys() -> None:
    errors: list[str] = []
    where = "config.pack"

    assert config_loader._coerce_value("max_total_bytes", None, errors, where=where) is None
    assert config_loader._coerce_value("report", None, errors, where=where) is None
    assert config_loader._coerce_value("out", "bundle.md", errors, where=where) == Path("bundle.md")
    assert config_loader._coerce_value("unsupported", "x", errors, where=where) == "x"

    assert errors == [f"{where}.unsupported: unsupported key"]


def test_load_command_config_accepts_policy_rules_tables(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                "",
                "[[pack.policy_rules]]",
                'rule_id = "scan-large"',
                'description = "Flag large files"',
                'stage = "scan"',
                "max_size_bytes = 100",
                'severity = "high"',
                'action = "deny"',
                "",
            ]
        ),
        encoding="utf-8",
    )

    values, _ = load_command_config("pack", root=tmp_path, config_path=config_path)
    assert values["policy_rules"] == [
        {
            "rule_id": "scan-large",
            "description": "Flag large files",
            "stage": "scan",
            "max_size_bytes": 100,
            "severity": "high",
            "action": "deny",
        }
    ]


def test_load_command_config_accepts_policy_pack_string(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                'policy_pack = "strict-privacy"',
                "",
            ]
        ),
        encoding="utf-8",
    )

    values, _ = load_command_config("pack", root=tmp_path, config_path=config_path)
    assert values["policy_pack"] == "strict-privacy"


def test_load_command_config_accepts_policy_enforcement_fields(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                "fail_on_policy_violation = true",
                'policy_fail_level = "high"',
                "",
            ]
        ),
        encoding="utf-8",
    )

    values, _ = load_command_config("pack", root=tmp_path, config_path=config_path)
    assert values["fail_on_policy_violation"] is True
    assert values["policy_fail_level"] == "high"


def test_load_command_config_accepts_policy_dry_run_fields(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                "policy_dry_run = true",
                'policy_output = "json"',
                "",
            ]
        ),
        encoding="utf-8",
    )

    values, _ = load_command_config("pack", root=tmp_path, config_path=config_path)
    assert values["policy_dry_run"] is True
    assert values["policy_output"] == "json"


def test_load_command_config_rejects_invalid_policy_fail_level(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                'policy_fail_level = "blocker"',
                "",
            ]
        ),
        encoding="utf-8",
    )

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("pack", root=tmp_path, config_path=config_path)

    assert "policy_fail_level: expected one of 'critical, high, low, medium'" in str(exc.value)


def test_load_command_config_rejects_invalid_policy_output(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                'policy_output = "yaml"',
                "",
            ]
        ),
        encoding="utf-8",
    )

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("pack", root=tmp_path, config_path=config_path)

    assert "policy_output: expected one of 'json, text'" in str(exc.value)


def test_load_command_config_rejects_non_bool_fail_on_policy_violation(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                'fail_on_policy_violation = "yes"',
                "",
            ]
        ),
        encoding="utf-8",
    )

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("pack", root=tmp_path, config_path=config_path)

    assert "fail_on_policy_violation: expected a boolean" in str(exc.value)


def test_load_command_config_rejects_non_bool_policy_dry_run(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                'policy_dry_run = "yes"',
                "",
            ]
        ),
        encoding="utf-8",
    )

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("pack", root=tmp_path, config_path=config_path)

    assert "policy_dry_run: expected a boolean" in str(exc.value)


def test_load_command_config_rejects_policy_rule_without_matchers(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                "",
                "[[pack.policy_rules]]",
                'rule_id = "bad-rule"',
                'description = "missing matcher"',
                "",
            ]
        ),
        encoding="utf-8",
    )

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("pack", root=tmp_path, config_path=config_path)

    assert "expected at least one matcher key" in str(exc.value)


def test_load_command_config_rejects_policy_rules_when_not_a_list(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                "policy_rules = 1",
                "",
            ]
        ),
        encoding="utf-8",
    )

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("pack", root=tmp_path, config_path=config_path)

    assert "expected a list of policy rule tables" in str(exc.value)


def test_load_command_config_rejects_non_string_policy_pack(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                "policy_pack = 1",
                "",
            ]
        ),
        encoding="utf-8",
    )

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("pack", root=tmp_path, config_path=config_path)

    assert "policy_pack: expected a non-empty string" in str(exc.value)


def test_load_command_config_rejects_policy_rule_entry_when_not_table(tmp_path: Path) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                "policy_rules = [1]",
                "",
            ]
        ),
        encoding="utf-8",
    )

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("pack", root=tmp_path, config_path=config_path)

    assert "expected a TOML table" in str(exc.value)


def test_load_command_config_rejects_policy_rule_invalid_fields_and_literals(
    tmp_path: Path,
) -> None:
    config_path = tmp_path / "foldermix.toml"
    config_path.write_text(
        "\n".join(
            [
                "[pack]",
                "",
                "[[pack.policy_rules]]",
                "rule_id = ''",
                "description = ''",
                "ext_in = 1",
                "max_size_bytes = -1",
                "severity = 1",
                "action = 'block'",
                "unknown = true",
                "",
            ]
        ),
        encoding="utf-8",
    )

    with pytest.raises(ConfigLoadError) as exc:
        load_command_config("pack", root=tmp_path, config_path=config_path)

    message = str(exc.value)
    assert "unknown key(s): unknown" in message
    assert "rule_id: expected a non-empty string" in message
    assert "description: expected a non-empty string" in message
    assert "ext_in: expected a list of strings" in message
    assert "max_size_bytes: expected a non-negative integer" in message
    assert "severity: expected a string" in message
    assert "action: expected one of deny, warn, got 'block'" in message


def test_extract_config_root_ignores_tool_table_without_foldermix(tmp_path: Path) -> None:
    parsed = {
        "tool": {"name": "demo"},
        "pack": {"workers": 2},
    }
    root = config_loader._extract_config_root(parsed, path=tmp_path / "foldermix.toml")
    assert root == parsed



from __future__ import annotations

from pathlib import Path

from .base import ConversionResult


class DocxFallbackConverter:
    def can_convert(self, ext: str) -> bool:
        try:
            import docx  # noqa: F401

            return ext.lower() == ".docx"
        except ImportError:
            return False

    def convert(self, path: Path, encoding: str = "utf-8") -> ConversionResult:
        import docx

        doc = docx.Document(str(path))
        paragraphs = [p.text for p in doc.paragraphs]
        return ConversionResult(
            content="\n\n".join(paragraphs),
            converter_name="python-docx",
            original_mime="application/vnd.openxmlformats-officedocument.wordprocessingml.document",
        )



from __future__ import annotations

import json
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import asdict
from pathlib import Path
from typing import cast

import typer
from rich.console import Console

from . import __version__
from .config import PackConfig
from .converters.base import ConverterRegistry
from .converters.pdf_fallback import PdfFallbackConverter
from .converters.registry import build_converter_registry
from .policy import PolicyEvaluator, normalize_policy_rules
from .policy_packs import combine_policy_rules
from .report import (
    ReportData,
    build_included_file_entry,
    build_policy_finding_counts,
    build_reason_code_counts,
    build_redaction_summary,
    build_skipped_file_entry,
    write_report,
)
from .scanner import FileRecord, SkipRecord, scan
from .utils import (
    drop_lines_containing,
    drop_lines_shorter_than,
    mtime_iso,
    sha256_file,
    utcnow_iso,
)
from .warning_taxonomy import normalize_warning_entries
from .writers.base import FileBundleItem, HeaderInfo
from .writers.jsonl_writer import JsonlWriter
from .writers.markdown_writer import MarkdownWriter
from .writers.xml_writer import XmlWriter

console = Console(stderr=True)
_POLICY_SEVERITY_ORDER: tuple[str, ...] = ("low", "medium", "high", "critical")
_POLICY_SEVERITY_RANK = {severity: idx for idx, severity in enumerate(_POLICY_SEVERITY_ORDER)}


def _count_failing_policy_findings(
    policy_findings: list[dict[str, object]], *, min_severity: str
) -> int:
    threshold_rank = _POLICY_SEVERITY_RANK[min_severity]
    failing_count = 0
    for finding in policy_findings:
        severity = finding.get("severity")
        if not isinstance(severity, str):
            continue
        severity_rank = _POLICY_SEVERITY_RANK.get(severity)
        if severity_rank is None:
            continue
        if severity_rank >= threshold_rank:
            failing_count += 1
    return failing_count


def _deny_policy_findings(policy_findings: list[dict[str, object]]) -> list[dict[str, object]]:
    return [
        finding
        for finding in policy_findings
        if isinstance(finding.get("action"), str) and finding.get("action") == "deny"
    ]


def _format_policy_severity_summary(by_severity: dict[str, int]) -> str:
    ordered_keys = [key for key in _POLICY_SEVERITY_ORDER if key in by_severity]
    ordered_keys.extend(sorted(key for key in by_severity if key not in _POLICY_SEVERITY_RANK))
    return ", ".join(f"{severity}={by_severity[severity]}" for severity in ordered_keys)


def _build_policy_stage_counts(policy_findings: list[dict[str, object]]) -> dict[str, int]:
    counts: dict[str, int] = {}
    for finding in policy_findings:
        stage = finding.get("stage")
        if isinstance(stage, str):
            counts[stage] = counts.get(stage, 0) + 1
    return dict(sorted(counts.items()))


def _build_affected_files(policy_findings: list[dict[str, object]]) -> list[dict[str, object]]:
    counts: dict[str, int] = {}
    for finding in policy_findings:
        path = finding.get("path")
        if isinstance(path, str):
            counts[path] = counts.get(path, 0) + 1
    return [{"path": path, "finding_count": counts[path]} for path in sorted(counts)]


def _sorted_policy_findings(policy_findings: list[dict[str, object]]) -> list[dict[str, object]]:
    def _finding_sort_key(
        finding: dict[str, object],
    ) -> tuple[int, str, str, str, str, str, str, str]:
        raw_path = finding.get("path")
        path = raw_path if isinstance(raw_path, str) else ""
        path_rank = 0 if isinstance(raw_path, str) else 1
        stage = finding.get("stage")
        rule_id = finding.get("rule_id")
        reason_code = finding.get("reason_code")
        severity = finding.get("severity")
        action = finding.get("action")
        message = finding.get("message")
        return (
            path_rank,
            path,
            stage if isinstance(stage, str) else "",
            rule_id if isinstance(rule_id, str) else "",
            reason_code if isinstance(reason_code, str) else "",
            severity if isinstance(severity, str) else "",
            action if isinstance(action, str) else "",
            message if isinstance(message, str) else "",
        )

    return sorted(policy_findings, key=_finding_sort_key)


def _build_policy_dry_run_payload(
    *, policy_findings: list[dict[str, object]], policy_counts: dict[str, object] | None
) -> dict[str, object]:
    counts = policy_counts or {
        "total": len(policy_findings),
        "by_severity": {},
        "by_action": {},
        "by_reason_code": {},
    }
    sorted_findings = _sorted_policy_findings(policy_findings)
    return {
        "schema_version": 1,
        "mode": "policy_dry_run",
        "finding_count": len(sorted_findings),
        "by_severity": counts["by_severity"],
        "by_action": counts["by_action"],
        "by_reason_code": counts["by_reason_code"],
        "by_stage": _build_policy_stage_counts(sorted_findings),
        "affected_files": _build_affected_files(sorted_findings),
        "non_file_finding_count": sum(
            1 for finding in sorted_findings if not isinstance(finding.get("path"), str)
        ),
        "findings": sorted_findings,
    }


def _print_policy_dry_run_text(payload: dict[str, object]) -> None:
    finding_count = cast(int, payload["finding_count"])
    by_severity = cast(dict[str, int], payload["by_severity"])
    affected_files = cast(list[dict[str, object]], payload["affected_files"])
    non_file_finding_count = cast(int, payload["non_file_finding_count"])

    console.print("[bold]Policy dry run complete.[/bold] No packed output written.")
    severity_summary = _format_policy_severity_summary(by_severity)
    suffix = f" ({severity_summary})" if severity_summary else ""
    console.print(f"[yellow]Policy findings:[/yellow] {finding_count}{suffix}")

    if affected_files:
        console.print(f"Affected files: {len(affected_files)}")
        for entry in affected_files:
            path = cast(str, entry["path"])
            finding_total = cast(int, entry["finding_count"])
            console.print(f"  [cyan]{path}[/cyan] ({finding_total} findings)")
    else:
        console.print("Affected files: 0")

    if non_file_finding_count:
        console.print(f"Non-file findings: {non_file_finding_count}")


def _write_report_if_requested(
    *,
    config: PackConfig,
    items: list[FileBundleItem],
    skipped: list[SkipRecord],
    total_bytes: int,
    policy_finding_entries: list[dict[str, object]],
    policy_counts: dict[str, object] | None,
) -> None:
    if not config.report:
        return

    included_files = [
        build_included_file_entry(
            path=item.relpath,
            size=item.size_bytes,
            ext=item.ext,
            truncated=item.truncated,
            redacted=item.redacted,
            redaction_event_count=item.redaction_event_count,
            redaction_categories=item.redaction_categories,
            warning_entries=item.warning_entries,
            redact_mode=config.redact,
        )
        for item in items
    ]
    skipped_files = [build_skipped_file_entry(path=r.relpath, reason=r.reason) for r in skipped]
    report_data = ReportData(
        included_count=len(items),
        skipped_count=len(skipped),
        total_bytes=total_bytes,
        included_files=included_files,
        skipped_files=skipped_files,
        policy_findings=policy_finding_entries,
        reason_code_counts=build_reason_code_counts(
            included_files=included_files,
            skipped_files=skipped_files,
        ),
        redaction_summary=build_redaction_summary(
            included_files=included_files,
            default_mode=config.redact,
        ),
        policy_finding_counts=(
            policy_counts
            if policy_counts is not None
            else build_policy_finding_counts(policy_findings=policy_finding_entries)
        ),
    )
    write_report(config.report, report_data)
    console.print(f"Report written to {config.report}")


def _enforce_policy_threshold_if_requested(
    *,
    enabled: bool,
    policy_findings: list[dict[str, object]],
    min_severity: str,
) -> None:
    if not enabled or not policy_findings:
        return

    deny_policy_findings = _deny_policy_findings(policy_findings)
    failing_count = _count_failing_policy_findings(
        deny_policy_findings,
        min_severity=min_severity,
    )
    if failing_count > 0:
        console.print(
            "[red]Policy enforcement failed:[/red] "
            f"{failing_count} deny finding(s) at or above "
            f"--policy-fail-level={min_severity!r}"
        )
        raise typer.Exit(code=4)


def _build_registry() -> ConverterRegistry:
    return build_converter_registry()


def _get_writer(fmt: str, include_toc: bool = True):
    if fmt == "xml":
        return XmlWriter()
    if fmt == "jsonl":
        return JsonlWriter()
    return MarkdownWriter(include_toc=include_toc)


def _convert_record(
    record: FileRecord,
    registry: ConverterRegistry,
    config: PackConfig,
) -> FileBundleItem:
    converter = registry.get_converter(record.ext)
    warnings: list[str] = []
    if record.ext == ".pdf" and config.pdf_ocr:
        pdf_converter = PdfFallbackConverter()
        if pdf_converter.can_convert(record.ext):
            converter = pdf_converter
        else:
            message = (
                "PDF OCR is enabled, but PDF/OCR dependencies are unavailable. "
                "Install the PDF/OCR extras (for example, 'pip install foldermix[ocr]') "
                "or disable --pdf-ocr."
            )
            if config.pdf_ocr_strict:
                raise RuntimeError(message)
            warnings.append(message)
    truncated = False
    redacted = False
    redaction_event_count = 0
    redaction_categories: list[str] = []

    if converter is None:
        content = f"[No converter available for {record.ext}]"
        converter_name = "none"
        original_mime = ""
    else:

        def run_convert(path: Path):
            if isinstance(converter, PdfFallbackConverter):
                return converter.convert(
                    path,
                    config.encoding,
                    enable_ocr=config.pdf_ocr,
                    ocr_strict=config.pdf_ocr_strict,
                )
            return converter.convert(path, config.encoding)

        try:
            if config.on_oversize == "truncate" and record.size > config.max_bytes:
                # Truncate: read first K and last K bytes
                k = config.max_bytes // 2
                with open(record.path, "rb") as f:
                    head = f.read(k)
                    f.seek(-min(k, record.size - k), 2)
                    tail = f.read(k)
                sep = b"\n\n... [TRUNCATED] ...\n\n"
                truncated_bytes = head + sep + tail
                tmp = record.path.parent / (record.path.name + ".truncated.tmp")
                try:
                    tmp.write_bytes(truncated_bytes)
                    result = run_convert(tmp)
                finally:
                    if tmp.exists():
                        tmp.unlink()
                truncated = True
            else:
                result = run_convert(record.path)

            content = result.content
            converter_name = result.converter_name
            original_mime = result.original_mime
            warnings.extend(result.warnings)

            if config.strip_frontmatter:
                from .utils import strip_yaml_frontmatter

                content = strip_yaml_frontmatter(content)

            if config.drop_line_containing:
                content = drop_lines_containing(content, config.drop_line_containing)

            content = drop_lines_shorter_than(content, config.min_line_length)

            if config.redact != "none":
                from .utils import apply_redaction_with_trace

                redacted_content, category_counts = apply_redaction_with_trace(
                    content, config.redact
                )
                redaction_event_count = sum(category_counts.values())
                redaction_categories = sorted(category_counts)
                redacted = redaction_event_count > 0
                content = redacted_content

            if config.line_ending == "crlf":
                content = content.replace("\r\n", "\n").replace("\n", "\r\n")

        except Exception as e:
            if config.continue_on_error:
                content = f"[Error converting file: {e}]"
                converter_name = "error"
                original_mime = ""
                warnings.append(str(e))
            else:
                raise

    sha256: str | None = None
    if config.include_sha256:
        try:
            sha256 = sha256_file(record.path)
        except OSError:
            pass

    return FileBundleItem(
        relpath=record.relpath,
        ext=record.ext,
        size_bytes=record.size,
        mtime=mtime_iso(record.path),
        sha256=sha256,
        content=content,
        converter_name=converter_name,
        original_mime=original_mime,
        warnings=warnings,
        warning_entries=normalize_warning_entries(warnings),
        truncated=truncated,
        redacted=redacted,
        redaction_mode=config.redact,
        redaction_event_count=redaction_event_count,
        redaction_categories=redaction_categories,
    )


def pack(config: PackConfig) -> None:
    """Main entry point for packing."""
    policy_evaluator: PolicyEvaluator | None = None
    try:
        raw_policy_rules = combine_policy_rules(
            policy_pack=config.policy_pack,
            policy_rules=config.policy_rules,
        )
    except ValueError as exc:
        console.print(f"[red]Invalid policy configuration:[/red] {exc}")
        raise typer.Exit(code=1) from exc

    if raw_policy_rules:
        try:
            policy_rules = normalize_policy_rules(raw_policy_rules)
        except ValueError as exc:
            console.print(f"[red]Invalid policy configuration:[/red] {exc}")
            raise typer.Exit(code=1) from exc
        policy_evaluator = PolicyEvaluator(policy_rules)

    policy_findings = []

    console.print(f"[bold]Scanning[/bold] {config.root} ...")
    included, skipped = scan(config)

    console.print(
        f"Found [green]{len(included)}[/green] files, [yellow]{len(skipped)}[/yellow] skipped"
    )

    if policy_evaluator is not None:
        for record in included:
            policy_findings.extend(
                policy_evaluator.evaluate_scan_included(
                    path=record.relpath,
                    ext=record.ext,
                    size_bytes=record.size,
                )
            )
        for skip in skipped:
            policy_findings.extend(
                policy_evaluator.evaluate_scan_skipped(
                    path=skip.relpath,
                    skip_reason=skip.reason,
                )
            )

    # Check limits
    if config.max_files is not None and len(included) > config.max_files:
        console.print(
            f"[red]Error:[/red] {len(included)} files exceeds --max-files={config.max_files}",
        )
        raise typer.Exit(code=3)

    if config.dry_run:
        for r in included:
            console.print(f"  [cyan]{r.relpath}[/cyan] ({r.size:,} bytes)")
        console.print(f"\n[bold]Dry run complete.[/bold] Would pack {len(included)} files.")
        return

    registry = _build_registry()
    writer = _get_writer(config.format, include_toc=config.include_toc)

    # Convert files
    items: list[FileBundleItem] = []
    errors: list[str] = []

    use_progress = config.progress
    try:
        import tqdm as _tqdm  # noqa: F401

        has_tqdm = True
    except ImportError:
        has_tqdm = False

    def convert_with_idx(record: FileRecord) -> tuple[FileRecord, FileBundleItem | Exception]:
        try:
            item = _convert_record(record, registry, config)
            return record, item
        except Exception as e:
            return record, e

    if use_progress and has_tqdm:
        from tqdm import tqdm

        with ThreadPoolExecutor(max_workers=config.workers) as executor:
            futures = {executor.submit(convert_with_idx, r): r for r in included}
            results = []
            for future in tqdm(as_completed(futures), total=len(included), desc="Converting"):
                results.append(future.result())
    else:
        with ThreadPoolExecutor(max_workers=config.workers) as executor:
            futures = {executor.submit(convert_with_idx, r): r for r in included}
            results = [f.result() for f in as_completed(futures)]

    # Re-sort to maintain deterministic order
    record_to_item: dict[str, FileBundleItem | Exception] = {}
    for record, result in results:
        record_to_item[record.relpath] = result

    for record in included:
        result = record_to_item[record.relpath]
        if isinstance(result, Exception):
            errors.append(f"{record.relpath}: {result}")
            console.print(f"[red]Error[/red] converting {record.relpath}: {result}")
        else:
            items.append(result)
            if policy_evaluator is not None:
                policy_findings.extend(
                    policy_evaluator.evaluate_converted(
                        path=result.relpath,
                        ext=result.ext,
                        size_bytes=result.size_bytes,
                        content=result.content,
                    )
                )

    if errors and not config.continue_on_error:
        console.print(
            f"[red]{len(errors)} conversion error(s). Use --continue-on-error to skip.[/red]"
        )
        raise typer.Exit(code=2)

    total_bytes = sum(i.size_bytes for i in items)

    if policy_evaluator is not None:
        policy_findings.extend(
            policy_evaluator.evaluate_pack_summary(
                file_count=len(items),
                total_bytes=total_bytes,
            )
        )

    policy_finding_entries = [asdict(finding) for finding in policy_findings]
    policy_counts: dict[str, object] | None = None
    if policy_finding_entries:
        policy_counts = build_policy_finding_counts(policy_findings=policy_finding_entries)
        if not config.policy_dry_run:
            by_severity = cast(dict[str, int], policy_counts["by_severity"])
            severity_summary = _format_policy_severity_summary(by_severity)
            suffix = f" ({severity_summary})" if severity_summary else ""
            console.print(f"[yellow]Policy findings:[/yellow] {policy_counts['total']}{suffix}")

    if config.max_total_bytes is not None and total_bytes > config.max_total_bytes:
        console.print(
            f"[red]Error:[/red] total {total_bytes:,} bytes exceeds --max-total-bytes={config.max_total_bytes:,}",
        )
        raise typer.Exit(code=3)

    if config.policy_dry_run:
        policy_payload = _build_policy_dry_run_payload(
            policy_findings=policy_finding_entries,
            policy_counts=policy_counts,
        )
        if config.policy_output == "json":
            typer.echo(json.dumps(policy_payload, indent=2, sort_keys=True))
        else:
            _print_policy_dry_run_text(policy_payload)

        _write_report_if_requested(
            config=config,
            items=items,
            skipped=skipped,
            total_bytes=total_bytes,
            policy_finding_entries=policy_finding_entries,
            policy_counts=policy_counts,
        )
        _enforce_policy_threshold_if_requested(
            enabled=config.fail_on_policy_violation,
            policy_findings=policy_finding_entries,
            min_severity=config.policy_fail_level,
        )
        return

    header = HeaderInfo(
        root=str(config.root),
        generated_at=utcnow_iso(),
        version=__version__,
        args={},
        file_count=len(items),
        total_bytes=total_bytes,
    )

    # Determine output path
    out_path = config.out
    if out_path is None:
        ext_map = {"md": "md", "xml": "xml", "jsonl": "jsonl"}
        from datetime import datetime, timezone

        ts = datetime.now(timezone.utc).strftime("%Y%m%d_%H%M%S")
        out_path = Path(f"foldermix_{ts}.{ext_map[config.format]}")

    console.print(f"Writing to [bold]{out_path}[/bold] ...")

    with open(out_path, "w", encoding="utf-8", newline="") as f:
        writer.write(f, header, items)

    console.print(f"[green]Done![/green] {len(items)} files, {total_bytes:,} bytes -> {out_path}")

    _write_report_if_requested(
        config=config,
        items=items,
        skipped=skipped,
        total_bytes=total_bytes,
        policy_finding_entries=policy_finding_entries,
        policy_counts=policy_counts,
    )
    _enforce_policy_threshold_if_requested(
        enabled=config.fail_on_policy_violation,
        policy_findings=policy_finding_entries,
        min_severity=config.policy_fail_level,
    )
